from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, Any, Literal

import numpy as np

from geff import _path
from geff.core_io._utils import (
    check_for_geff,
    construct_var_len_props,
    delete_geff,
    remove_tilde,
    setup_zarr_group,
)
from geff.validate.structure import validate_structure
from geff_spec.utils import (
    add_or_update_props_metadata,
    compute_and_add_axis_min_max,
    create_props_metadata,
)

from ._serialization import serialize_vlen_property_data

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from zarr.storage import StoreLike

    from geff._typing import PropDictNpArray
    from geff_spec import GeffMetadata, PropMetadata


def write_dicts(
    geff_store: StoreLike,
    node_data: Iterable[tuple[Any, dict[str, Any]]],
    edge_data: Iterable[tuple[Any, dict[str, Any]]],
    node_prop_names: Sequence[str],
    edge_prop_names: Sequence[str],
    metadata: GeffMetadata,
    zarr_format: Literal[2, 3] = 2,
    structure_validation: bool = True,
) -> None:
    """Write a dict-like graph representation to geff

    Args:
        geff_store (zarr.storage.StoreLike): The path/str to the geff zarr, or the store
            itself. Opens in append mode, so will only overwrite geff-controlled groups.
        node_data (Sequence[tuple[Any, dict[str, Any]]]): A sequence of tuples with
            node_ids and node_data, where node_data is a dictionary from str names
            to any values.
        edge_data (Sequence[tuple[Any, dict[str, Any]]]): A sequence of tuples with
            edge_ids and edge_data, where edge_data is a dictionary from str names
            to any values.
        node_prop_names (Sequence[str]): A list of node properties to include in the
            geff
        edge_prop_names (Sequence[str]): A list of edge properties to include in the
            geff
        metadata (GeffMetadata): The core metadata to write. Node/edge properties and
            axis min and maxes will be overwritten.
        zarr_format (Literal[2, 3]): The zarr specification to use when writing the zarr.
            Defaults to 2.
        structure_validation (bool): If True, runs structural validation and does not write
            a geff that is invalid. Defaults to True.

    Raises:
        ValueError: If the position prop is given and is not present on all nodes.
    """

    geff_store = remove_tilde(geff_store)

    node_data = list(node_data)
    edge_data = list(edge_data)
    node_ids = [idx for idx, _ in node_data]
    edge_ids = [idx for idx, _ in edge_data]

    if len(node_ids) > 0:
        nodes_arr = np.asarray(node_ids)
        # Check if we can cast to uint
        if any(nodes_arr < 0):
            raise ValueError("Cannot write a geff with node ids that are negative")
        if not np.issubdtype(nodes_arr.dtype, np.integer):
            warnings.warn(
                f"Node ids with dtype {nodes_arr.dtype} are being cast to uint", stacklevel=2
            )
        nodes_arr = nodes_arr.astype("uint")
    else:
        nodes_arr = np.empty((0,), dtype=np.uint64)

    if len(edge_ids) > 0:
        edges_arr = np.asarray(edge_ids, dtype=nodes_arr.dtype)
    else:
        edges_arr = np.empty((0, 2), dtype=nodes_arr.dtype)

    node_props_dict = dict_props_to_arr(node_data, node_prop_names)

    edge_props_dict = dict_props_to_arr(edge_data, edge_prop_names)
    write_arrays(
        geff_store,
        nodes_arr,
        node_props_dict,
        edges_arr,
        edge_props_dict,
        metadata,
        zarr_format=zarr_format,
        structure_validation=structure_validation,
    )


def _determine_default_value(data: Sequence[tuple[Any, dict[str, Any]]], prop_name: str) -> Any:
    """Determine default value to fill in missing values

    Find the first non-missing value and then uses the following heuristics:
    - Native python numerical types (int, float) -> 0
    - Native python string -> ""
    - Otherwise, return the  value, which is definitely the right type and
    shape, but is potentially both confusing and inefficient. Should reconsider in
    the future.

    If there are no non-missing values, warns and then returns 0.

    Args:
        data (Sequence[tuple[Any, dict[str, Any]]]): A sequence of elements and a dictionary
            holding the properties of that element
        prop_name (str): The property to get the default value for

    Returns:
        Any: A value to use as the default that is the same dtype and shape as the rest
            of the values, for casting to a numpy array without errors.
    """
    for _, data_dict in data:
        # find first non-missing value
        if prop_name in data_dict:
            value = data_dict[prop_name]
            if isinstance(value, int | float):
                return 0
            elif isinstance(value, str):
                return ""
            else:
                return value
    warnings.warn(
        f"Property {prop_name} is not present on any graph elements. Using 0 as the default.",
        stacklevel=2,
    )
    return 0


def dict_props_to_arr(
    data: Sequence[tuple[Any, dict[str, Any]]],
    prop_names: Sequence[str],
) -> dict[str, PropDictNpArray]:
    """Convert dict-like properties to values and missing array representation.

    Note: The order of the sequence of data should be the same as that used to write
    the ids, or this will not work properly.

    Args:
        data (Sequence[tuple[Any, dict[str, Any]]]): A sequence of elements and a dictionary
            holding the properties of that element
        prop_names (str): The properties to include in the dictionary of property arrays.

    Returns:
        dict[PropDictNpArray]: A dictionary from property names
            to a tuple of (value, missing) arrays, where the missing array can be None.
    """
    props_dict: dict[str, PropDictNpArray] = {}
    for name in prop_names:
        values = []
        missing = []
        # iterate over the data and checks for missing content
        missing_any = False
        # to ensure valid dtype of missing, take first non-missing value
        default_val = None
        for _, data_dict in data:
            if name in data_dict:
                values.append(data_dict[name])
                missing.append(False)
            else:
                if default_val is None:
                    default_val = _determine_default_value(data, name)
                values.append(default_val)
                missing.append(True)
                missing_any = True
        # try casting the list of values to a numpy array
        try:
            values_arr = np.asarray(values)
        # catch a value error which will happen if we have elements that are different shapes
        except ValueError:
            # try to construct variable length properties - will raise an error if internal
            # dtypes are not compatible (e.g. floats and strings)
            values_arr = construct_var_len_props(values)["values"]
        missing_arr = np.asarray(missing, dtype=bool) if missing_any else None
        props_dict[name] = {"missing": missing_arr, "values": values_arr}
    return props_dict


def write_arrays(
    geff_store: StoreLike,
    node_ids: np.ndarray,
    node_props: dict[str, PropDictNpArray] | None,
    edge_ids: np.ndarray,
    edge_props: dict[str, PropDictNpArray] | None,
    metadata: GeffMetadata,
    node_props_unsquish: dict[str, list[str]] | None = None,
    edge_props_unsquish: dict[str, list[str]] | None = None,
    zarr_format: Literal[2, 3] = 2,
    structure_validation: bool = True,
    overwrite: bool = False,
) -> None:
    """Write a geff file from already constructed arrays of node and edge ids and props

    Currently does not do any validation that the arrays are valid, but could be added
    as an optional flag.
    Adds the PropMetadata for the nodes and edges, if not provided.

    Args:
        geff_store (str | Path | zarr store): The path/str to the geff zarr, or the store
            itself. Opens in append mode, so will only overwrite geff-controlled groups.
        node_ids (np.ndarray): An array containing the node ids. Must have same dtype as
            edge_ids.
        node_props (dict[str, PropDictNpArray] | None): A dictionary
            from node property names to (values, missing) arrays, which should have same
            length as node_ids.
        edge_ids (np.ndarray): An array containing the edge ids. Must have same dtype
            as node_ids.
        edge_props (dict[str, PropDictNpArray] | None): A dictionary
            from edge property names to (values, missing) arrays, which should have same
            length as edge_ids.
        metadata (GeffMetadata): The metadata of the graph.
        zarr_format (Literal[2, 3]): The zarr specification to use when writing the zarr.
            Defaults to 2.
        node_props_unsquish (dict[str, list[str]] | None): a dictionary
            indicication how to "unsquish" a property into individual scalars
            (e.g.: `{"pos": ["z", "y", "x"]}` will store the position property
            as three individual properties called "z", "y", and "x".
        edge_props_unsquish (dict[str, list[str]] | None): a dictionary
            indicication how to "unsquish" a property into individual scalars
            (e.g.: `{"pos": ["z", "y", "x"]}` will store the position property
            as three individual properties called "z", "y", and "x".
        structure_validation (bool): If True, runs structural validation and does not write
            a geff that is invalid. Defaults to True.
        overwrite (bool): If True, deletes any existing geff and writes a new geff.
            Defaults to False.

    Raises:
        FileExistsError: If a geff already exists in `geff_store`
    """
    geff_store = remove_tilde(geff_store)

    # Check for an existing geff
    if check_for_geff(geff_store):
        if overwrite:
            delete_geff(geff_store, zarr_format=zarr_format)
        else:
            raise FileExistsError(
                "Found an existing geff present in `geff_store`. "
                "Please use `overwrite=True` or provide an alternative "
                "`geff_store` to write to."
            )

    write_id_arrays(geff_store, node_ids, edge_ids, zarr_format=zarr_format)
    if node_props is not None:
        node_meta = write_props_arrays(
            geff_store, _path.NODES, node_props, node_props_unsquish, zarr_format=zarr_format
        )
    else:
        node_meta = []
    if edge_props is not None:
        edge_meta = write_props_arrays(
            geff_store, _path.EDGES, edge_props, edge_props_unsquish, zarr_format=zarr_format
        )
    else:
        edge_meta = []

    metadata = add_or_update_props_metadata(metadata, node_meta, "node")
    metadata = add_or_update_props_metadata(metadata, edge_meta, "edge")
    if node_props is not None:
        metadata = compute_and_add_axis_min_max(metadata, node_props)
    metadata.write(geff_store)

    if structure_validation:
        try:
            validate_structure(geff_store)
        except ValueError as e:
            message = "\nCannot write invalid geff."
            try:
                delete_geff(geff_store, zarr_format=zarr_format)
            except:  # noqa: E722
                message = (
                    "\nWritten geff is invalid, but cannot be deleted automatically. "
                    "Please delete manually."
                )
            raise ValueError(e.args[0] + message) from e


def write_id_arrays(
    geff_store: StoreLike,
    node_ids: np.ndarray,
    edge_ids: np.ndarray,
    zarr_format: Literal[2, 3] = 2,
) -> None:
    """Writes a set of node ids and edge ids to a geff group.

    Args:
        geff_store (str | Path | zarr store): path/str to geff group, or the store itself,
            to write the nodes/ids and edges/ids into
        node_ids (np.ndarray): an array of strings or ints with shape (N,)
        edge_ids (np.ndarray): an array with same type as node_ds and shape (N, 2)
        zarr_format (Literal[2, 3]): The zarr specification to use when writing the zarr.
            Defaults to 2.
    Raises:
        TypeError if node_ids and edge_ids have different types, or if either are float
    """
    geff_store = remove_tilde(geff_store)

    if node_ids.dtype != edge_ids.dtype:
        raise TypeError(
            f"Node ids and edge ids must have same dtype: {node_ids.dtype=}, {edge_ids.dtype=}"
        )
    if not np.issubdtype(node_ids.dtype, np.integer):
        raise TypeError(
            f"Node ids and edge ids must have int dtype: {node_ids.dtype=}, {edge_ids.dtype=}"
        )

    geff_root = setup_zarr_group(geff_store, zarr_format)
    geff_root[_path.NODE_IDS] = node_ids
    geff_root[_path.EDGE_IDS] = edge_ids


def write_props_arrays(
    geff_store: StoreLike,
    group: Literal["nodes", "edges"],
    props: dict[str, PropDictNpArray],
    props_unsquish: dict[str, list[str]] | None = None,
    zarr_format: Literal[2, 3] = 2,
) -> Sequence[PropMetadata]:
    """Writes a set of properties to a geff nodes or edges group.

    Can be used to add new properties if they don't already exist.

    Args:
        geff_store (str | Path | zarr store): The path/str to the geff zarr, or the store
            itself. Opens in append mode, so will only overwrite geff-controlled groups.
        group (Literal["nodes", "edges"]): "nodes" or "edges"
        props (dict[str, PropDictNpArray]): a dictionary from
            attr name to (attr_values, attr_missing) arrays.
        props_unsquish (dict[str, list[str]] | None): a dictionary indicication
            how to "unsquish" a property into individual scalars (e.g.:
            `{"pos": ["z", "y", "x"]}` will store the position property as
            three individual properties called "z", "y", and "x".
        zarr_format (Literal[2, 3]): The zarr specification to use when writing the zarr.
            Defaults to 2.

    Returns:
        PropMetadata: The property metadata for each of the property arrays
    Raises:
        ValueError: If the group is not a 'nodes' or 'edges' group.
    TODO: validate attrs length based on group ids shape?
    """

    geff_store = remove_tilde(geff_store)

    if group not in [_path.NODES, _path.EDGES]:
        raise ValueError(f"Group must be a {_path.NODES!r} or {_path.EDGES!r} group. Found {group}")

    if props_unsquish is not None:
        for name, replace_names in props_unsquish.items():
            values = props[name]["values"]
            missing = props[name]["missing"]
            if (not len(values.shape) == 2) or np.issubdtype(values.dtype, np.object_):
                raise ValueError(
                    "Can only unsquish 2D array properties that are not variable length."
                )

            replace_arrays: dict[str, PropDictNpArray]
            for i, replace_name in enumerate(replace_names):
                replace_arrays = {
                    replace_name: {
                        "values": values[:, i],
                        "missing": None if missing is None else missing,
                    }
                }
                props.update(replace_arrays)
            del props[name]

    geff_root = setup_zarr_group(geff_store, zarr_format)
    props_group = geff_root.require_group(f"{group}/{_path.PROPS}")
    metadata = []
    for prop, prop_dict in props.items():
        prop_metadata = create_props_metadata(prop, prop_dict)
        metadata.append(prop_metadata)

        if prop_metadata.varlength:
            values, missing, data = serialize_vlen_property_data(prop_dict)
        else:
            values = prop_dict["values"]
            missing = prop_dict["missing"]
            data = None

        prop_group = props_group.create_group(prop)
        prop_group[_path.VALUES] = values
        if missing is not None:
            prop_group[_path.MISSING] = missing
        if data is not None:
            prop_group[_path.DATA] = data

    return metadata



from pathlib import Path
from typing import TYPE_CHECKING, Annotated, cast

import typer

from geff_spec import GeffMetadata

from . import validate_structure

if TYPE_CHECKING:
    from typing import Literal

app = typer.Typer(help="GEFF Command Line Interface")


@app.command()
def validate(
    input_path: str = typer.Argument(
        ..., help="Path to the GEFF folder, e.g. data.zarr/tracks.geff"
    ),
) -> None:
    """Validate a GEFF file."""
    validate_structure(input_path)


@app.command()
def info(
    input_path: str = typer.Argument(
        ..., help="Path to the GEFF folder, e.g. data.zarr/tracks.geff"
    ),
) -> None:
    """Display information about a GEFF file."""
    metadata = GeffMetadata.read(input_path)
    print(metadata.model_dump_json(indent=2))  # noqa


@app.command()
def convert_ctc(
    ctc_path: Annotated[
        Path,
        typer.Argument(
            help="The path to the directory containing ctc tracks (man_track.txt or res_track.txt),"
            " e.g. data/01_GT/TRA or data/01_RES",
            show_default=False,
        ),
    ],
    geff_path: Annotated[
        Path,
        typer.Argument(
            help="Path to save the output geff, including the geff directory name "
            "(eg. ~/folder/folder/data.zarr/tracks.geff)",
            show_default=False,
        ),
    ],
    segm_path: Annotated[
        Path | None,
        typer.Option(
            help="The path to export the segmentation file, if not provided, it won't be exported."
        ),
    ] = None,
    input_image_dir: Annotated[
        Path | None,
        typer.Option(
            help="The path to the input image directory. If not provided, it won't be exported."
        ),
    ] = None,
    output_image_path: Annotated[
        Path | None,
        typer.Option(
            help="The path to export the image file, if not provided, it won't be exported."
        ),
    ] = None,
    tczyx: Annotated[
        bool,
        typer.Option(
            help="Expand data to make it (T, C, Z, Y, X) otherwise it's (T,) + Frame shape."
        ),
    ] = False,
    overwrite: Annotated[
        bool,
        typer.Option(help="Whether to overwrite the GEFF file if it already exists."),
    ] = False,
    zarr_format: Annotated[int, typer.Option(help="The version of zarr to write.")] = 2,
    # because of Typer not supporting Literal types
) -> None:
    """
    Convert a CTC data directory to a GEFF file.
    """
    from geff.convert import from_ctc_to_geff, ctc_tiffs_to_zarr  # noqa: I001 import at call time to avoid optional dependency issues

    if (input_image_dir is not None and output_image_path is None) or (
        input_image_dir is None and output_image_path is not None
    ):
        raise ValueError("'input_image_dir' and 'output_image_path' must be provided together")

    from_ctc_to_geff(
        ctc_path=ctc_path,
        geff_path=geff_path,
        segmentation_store=segm_path,
        tczyx=tczyx,
        overwrite=overwrite,
        zarr_format=cast("Literal[2, 3]", zarr_format),
    )

    if input_image_dir is not None:
        ctc_tiffs_to_zarr(
            ctc_path=input_image_dir,
            output_store=output_image_path,
            ctzyx=tczyx,
            overwrite=overwrite,
            zarr_format=cast("Literal[2, 3]", zarr_format),
        )


@app.command()
def convert_trackmate_xml(
    xml_path: Annotated[
        Path,
        typer.Argument(help="The path to the TrackMate XML file", show_default=False),
    ],
    geff_path: Annotated[
        Path,
        typer.Argument(
            help="Path to save the output geff, including the geff directory name"
            " (eg. ~/folder/folder/data.zarr/tracks.geff)",
            show_default=False,
        ),
    ],
    discard_filtered_spots: Annotated[
        bool,
        typer.Option(help="True to discard the spots filtered out in TrackMate, False otherwise."),
    ] = False,
    discard_filtered_tracks: Annotated[
        bool,
        typer.Option(help="True to discard the tracks filtered out in TrackMate, False otherwise."),
    ] = False,
    overwrite: Annotated[
        bool,
        typer.Option(help="Whether to overwrite the GEFF file if it already exists."),
    ] = False,
    zarr_format: Annotated[int, typer.Option(help="The version of zarr to write.")] = 2,
    # because of Typer not supporting Literal types
) -> None:
    """
    Convert a TrackMate XML file to a GEFF file.
    """
    from geff.convert import from_trackmate_xml_to_geff

    from_trackmate_xml_to_geff(
        xml_path=xml_path,
        geff_path=geff_path,
        discard_filtered_spots=discard_filtered_spots,
        discard_filtered_tracks=discard_filtered_tracks,
        overwrite=overwrite,
        zarr_format=cast("Literal[2, 3]", zarr_format),
    )


@app.command()
def convert_to_csv(
    store: Annotated[Path, typer.Argument(help="Path to geff group to convert")],
    outpath: Annotated[
        Path,
        typer.Argument(
            help="Path to save output csvs. Any file extension will"
            " be stripped and replaced with '-nodes.csv' and '-edges.csv'"
        ),
    ],
) -> None:
    from geff.convert import geff_to_csv

    geff_to_csv(store=store, outpath=outpath)


if __name__ == "__main__":
    app()



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal

import networkx as nx
import numpy as np

from geff.core_io import write_dicts
from geff_spec.utils import create_or_update_metadata, update_metadata_axes

from ._backend_protocol import Backend
from ._graph_adapter import GraphAdapter

if TYPE_CHECKING:
    from collections.abc import Sequence

    from numpy.typing import NDArray
    from zarr.storage import StoreLike

    from geff._typing import PropDictNpArray
    from geff_spec import AxisType, GeffMetadata

import logging

logger = logging.getLogger(__name__)


def _set_property_values(
    graph: nx.Graph,
    ids: NDArray[Any],
    name: str,
    prop_dict: PropDictNpArray,
    nodes: bool = True,
) -> None:
    """Add properties in-place to a networkx graph's
    nodes or edges by creating attributes on the nodes/edges

    Args:
        graph (nx.DiGraph): The networkx graph, already populated with nodes or edges,
            that needs properties added
        ids (np.ndarray): Node or edge ids from Geff. If nodes, 1D. If edges, 2D.
        name (str): The name of the property.
        prop_dict (PropDict[np.ndarray]): A dictionary containing a "values" key with
            an array of values and an optional "missing" key for missing values.
        nodes (bool, optional): If True, extract and set node properties.  If False,
            extract and set edge properties. Defaults to True.
    """
    values = prop_dict["values"]
    varlength = np.issubdtype(values.dtype, np.object_)
    for idx in range(len(ids)):
        _id = ids[idx]
        val = prop_dict["values"][idx]
        # If property is sparse and missing for this node, skip setting property
        ignore = prop_dict["missing"][idx] if prop_dict["missing"] is not None else False
        if not ignore:
            # if varlength, store numpy arrays on graph. Otherwise,
            # Get either individual item or list instead of setting with np.array
            value = val if varlength else val.tolist()
            if nodes:
                graph.nodes[_id.item()][name] = value
            else:
                source, target = _id.tolist()
                graph.edges[source, target][name] = value


# NOTE: see _api_wrapper.py read/write/construct for docs
class NxBackend(Backend):
    @property
    def GRAPH_TYPES(self) -> tuple[type[nx.Graph], type[nx.DiGraph]]:
        return nx.Graph, nx.DiGraph

    @staticmethod
    def construct(
        metadata: GeffMetadata,
        node_ids: NDArray[Any],
        edge_ids: NDArray[Any],
        node_props: dict[str, PropDictNpArray],
        edge_props: dict[str, PropDictNpArray],
    ) -> nx.Graph | nx.DiGraph:
        graph = nx.DiGraph() if metadata.directed else nx.Graph()

        graph.add_nodes_from(node_ids.tolist())
        for name, prop_dict in node_props.items():
            _set_property_values(graph, node_ids, name, prop_dict, nodes=True)

        graph.add_edges_from(edge_ids.tolist())
        for name, prop_dict in edge_props.items():
            _set_property_values(graph, edge_ids, name, prop_dict, nodes=False)

        return graph

    @staticmethod
    def write(
        graph: nx.Graph | nx.DiGraph,
        store: StoreLike,
        metadata: GeffMetadata | None = None,
        axis_names: list[str] | None = None,
        axis_units: list[str | None] | None = None,
        axis_types: list[AxisType | None] | None = None,
        axis_scales: list[float | None] | None = None,
        scaled_units: list[str | None] | None = None,
        axis_offset: list[float | None] | None = None,
        zarr_format: Literal[2, 3] = 2,
        structure_validation: bool = True,
    ) -> None:
        directed = isinstance(graph, nx.DiGraph)
        metadata = create_or_update_metadata(metadata=metadata, is_directed=directed)
        if axis_names is not None:
            metadata = update_metadata_axes(
                metadata, axis_names, axis_units, axis_types, axis_scales, scaled_units, axis_offset
            )
        node_props = list({k for _, data in graph.nodes(data=True) for k in data})

        edge_data = [((u, v), data) for u, v, data in graph.edges(data=True)]
        edge_props = list({k for _, _, data in graph.edges(data=True) for k in data})
        write_dicts(
            store,
            graph.nodes(data=True),
            edge_data,
            node_props,
            edge_props,
            metadata,
            zarr_format=zarr_format,
            structure_validation=structure_validation,
        )

    @staticmethod
    def graph_adapter(graph: Any) -> NxGraphAdapter:
        return NxGraphAdapter(graph)


class NxGraphAdapter(GraphAdapter):
    def __init__(self, graph: nx.Graph | nx.DiGraph) -> None:
        self.graph = graph

    def get_node_ids(self) -> Sequence[int]:
        return list(self.graph.nodes)

    def get_edge_ids(self) -> Sequence[tuple[int, int]]:
        return list(self.graph.edges)

    def has_node_prop(self, name: str, node: int, metadata: GeffMetadata) -> bool:
        return name in self.graph.nodes[node]

    def get_node_prop(
        self,
        name: str,
        node: int,
        metadata: GeffMetadata,
    ) -> Any:
        return self.graph.nodes[node][name]

    def has_edge_prop(self, name: str, edge: tuple[int, int], metadata: GeffMetadata) -> bool:
        return name in self.graph.edges[edge]

    def get_edge_prop(
        self,
        name: str,
        edge: tuple[int, int],
        metadata: GeffMetadata,
    ) -> Any:
        return self.graph.edges[edge][name]



from typing import TYPE_CHECKING, get_args

import networkx as nx
import numpy as np
import pytest
import zarr

from geff import construct, read, write
from geff._graph_libs._api_wrapper import SupportedBackend, get_backend
from geff._graph_libs._backend_protocol import GraphAdapter
from geff._graph_libs._networkx import NxBackend
from geff._typing import InMemoryGeff
from geff.core_io import read_to_memory
from geff.testing.data import (
    create_empty_geff,
    create_mock_geff,
    create_simple_2d_geff,
    create_simple_3d_geff,
)

if TYPE_CHECKING:
    from geff._graph_libs._backend_protocol import Backend


rx = pytest.importorskip("rustworkx")
sg = pytest.importorskip("spatial_graph")


# assert that all the data in the graph are equal to those in the memory geff it was created from
def _assert_graph_equal_to_geff(
    graph_adapter: GraphAdapter,
    memory_geff: InMemoryGeff,
):
    metadata = memory_geff["metadata"]

    # nodes and edges correct
    assert {*graph_adapter.get_node_ids()} == {*memory_geff["node_ids"].tolist()}
    assert {*graph_adapter.get_edge_ids()} == {
        *[tuple(edges) for edges in memory_geff["edge_ids"].tolist()]
    }

    for name, data in memory_geff["node_props"].items():
        values = data["values"]
        missing = data["missing"]
        if missing is None:
            missing = np.zeros(shape=(values.shape[0],), dtype=bool)
        nodes = memory_geff["node_ids"]
        for node, expected_val, expected_missing in zip(nodes, values, missing, strict=True):
            actual_missing = not graph_adapter.has_node_prop(name, node, metadata)
            assert actual_missing == expected_missing

            if not expected_missing:
                actual_val = graph_adapter.get_node_prop(name, node, metadata=metadata)
                if isinstance(actual_val, np.ndarray):
                    np.testing.assert_array_equal(expected_val, actual_val)
                else:
                    assert expected_val == actual_val

    # check edge properties are correct
    for name, data in memory_geff["edge_props"].items():
        values = data["values"]
        missing = data["missing"]
        if missing is None:
            missing = np.zeros(shape=(values.shape[0],), dtype=bool)

        edges = memory_geff["edge_ids"]
        for edge, expected_val, expected_missing in zip(edges, values, missing, strict=True):
            actual_missing = not graph_adapter.has_edge_prop(name, edge, metadata)
            assert actual_missing == expected_missing
            if not expected_missing:
                actual_val = graph_adapter.get_edge_prop(name, edge, metadata)
                if isinstance(actual_val, np.ndarray):
                    np.testing.assert_array_equal(expected_val, actual_val)
                else:
                    assert expected_val == actual_val


PROP_DTYPES = ["float64", "uint8", "float32", "int16"]
ID_DTYPES = ["int8", "uint16"]
AXIS_DTYPES = {"position": "double", "time": "int"}


@pytest.mark.parametrize("node_id_dtype", ID_DTYPES, ids=["nid_" + dtype for dtype in ID_DTYPES])
@pytest.mark.parametrize("directed", [True, False], ids=["direct", "!direct"])
@pytest.mark.parametrize("include_t", [True, False], ids=["t", "!t"])
@pytest.mark.parametrize("include_spatial", [True, False], ids=["xyz", "!xyz"])
@pytest.mark.parametrize("backend", get_args(SupportedBackend))
class Test_api_wrapper:
    def test_read(
        self,
        node_id_dtype,
        directed,
        include_t,
        include_spatial,
        backend,
    ) -> None:
        if include_spatial is False and backend == "spatial-graph":
            pytest.skip("Non-spatial graphs not supported by spatial-graph")
        backend_module: Backend = get_backend(backend)

        extra_props = {dtype: dtype for dtype in PROP_DTYPES}
        if backend != "spatial-graph":
            extra_props["str"] = "str"

        store, memory_geff = create_mock_geff(
            node_id_dtype,
            AXIS_DTYPES,
            extra_node_props=extra_props,
            extra_edge_props=extra_props,
            directed=directed,
            include_t=include_t,
            include_x=include_spatial,
            include_y=include_spatial,
            include_z=include_spatial,
            include_varlength=backend != "spatial-graph",
            include_missing=backend != "spatial-graph",
        )

        graph, _metadata = read(store, backend=backend)
        graph_adapter = backend_module.graph_adapter(graph)

        _assert_graph_equal_to_geff(graph_adapter, memory_geff)

    def test_construct(
        self,
        node_id_dtype,
        directed,
        include_t,
        include_spatial,
        backend,
    ) -> None:
        backend_module: Backend = get_backend(backend)

        extra_props = {dtype: dtype for dtype in PROP_DTYPES}
        if backend != "spatial-graph":
            extra_props["str"] = "str"

        store, memory_geff = create_mock_geff(
            node_id_dtype,
            AXIS_DTYPES,
            extra_node_props=extra_props,
            extra_edge_props=extra_props,
            directed=directed,
            include_t=include_t,
            include_x=include_spatial,
            include_y=include_spatial,
            include_z=include_spatial,
            include_varlength=backend != "spatial-graph",
            include_missing=backend != "spatial-graph",
        )

        in_memory_geff = read_to_memory(store)
        if all([include_spatial is False, include_t is False, backend == "spatial-graph"]):
            with pytest.raises(
                ValueError,
                match="Cannot construct a non-empty SpatialGraph from a geff without axes",
            ):
                graph = construct(**in_memory_geff, backend=backend)
        else:
            graph = construct(**in_memory_geff, backend=backend)
            graph_adapter = backend_module.graph_adapter(graph)

            _assert_graph_equal_to_geff(graph_adapter, memory_geff)

    def test_write(
        self,
        tmp_path,
        node_id_dtype,
        directed,
        include_t,
        include_spatial,
        backend,
    ) -> None:
        if include_spatial is False and backend == "spatial-graph":
            pytest.skip("Non-spatial graphs not supported by spatial-graph")
        backend_module: Backend = get_backend(backend)

        extra_props = {dtype: dtype for dtype in PROP_DTYPES}
        if backend != "spatial-graph":
            extra_props["str"] = "str"

        _store, memory_geff = create_mock_geff(
            node_id_dtype,
            AXIS_DTYPES,
            extra_node_props=extra_props,
            extra_edge_props=extra_props,
            directed=directed,
            include_t=include_t,
            include_x=include_spatial,
            include_y=include_spatial,
            include_z=include_spatial,
            include_varlength=backend != "spatial-graph",
            include_missing=backend != "spatial-graph",
        )

        # this will create a graph instance of the backend type
        original_graph = backend_module.construct(**memory_geff)

        # Add extra values to metadata
        metadata = memory_geff["metadata"]
        metadata.extra = {"foo": "bar", "bar": {"baz": "qux"}}

        # write with unified write function
        path_store = tmp_path / "test_path.zarr"
        write(original_graph, path_store, memory_geff["metadata"])

        # read with the NxBackend to see if the graph is the same
        graph, metadata = NxBackend.read(path_store)
        graph_adapter = NxBackend.graph_adapter(graph)

        _assert_graph_equal_to_geff(graph_adapter, memory_geff)
        assert metadata.extra["foo"] == "bar"
        assert metadata.extra["bar"]["baz"] == "qux"


@pytest.mark.parametrize("backend", get_args(SupportedBackend))
class Test_api_wrapper_simple:  # tests that only need backend parametrization
    def test_write_axis_lists_override_metadata(self, tmp_path, backend):
        backend_module: Backend = get_backend(backend)
        _store, memory_geff = create_simple_3d_geff()

        # this will create a graph instance of the backend type
        original_graph = backend_module.construct(**memory_geff)

        # write with unified write function
        path_store = tmp_path / "test_path.zarr"
        axis_names = ["x", "y"]
        axis_units = ["meter", "meter"]
        axis_types = ["space", "space"]
        if backend == "spatial-graph":
            # Cannot change the number of axes in sg graph without modifying position attribute
            with pytest.raises(
                ValueError,
                match=r"Cannot write a SpatialGraph with ndims .* "
                "and a different number of axes (.*)",
            ):
                backend_module.write(
                    original_graph,
                    path_store,
                    metadata=memory_geff["metadata"],
                    axis_names=axis_names,
                    axis_units=axis_units,
                    axis_types=axis_types,
                )
        else:
            backend_module.write(
                original_graph,
                path_store,
                metadata=memory_geff["metadata"],
                axis_names=axis_names,
                axis_units=axis_units,
                axis_types=axis_types,
            )

            new_mem_geff = read_to_memory(path_store)
            metadata = new_mem_geff["metadata"]
            assert metadata.axes is not None
            assert len(metadata.axes) == 2
            assert axis_names == [axis.name for axis in metadata.axes]
            assert axis_units == [axis.unit for axis in metadata.axes]
            assert axis_types == [axis.type for axis in metadata.axes]

    def test_write_read_different_stores(self, tmp_path, backend):
        stores = [
            tmp_path / "test_path.zarr",  # Path object
            str(tmp_path / "test_string.zarr"),  # string path
            zarr.storage.MemoryStore(),
        ]

        backend_module: Backend = get_backend(backend)
        _, memory_geff = create_simple_3d_geff()

        # this will create a graph instance of the backend type
        original_graph = backend_module.construct(**memory_geff)
        adpt_og_graph = backend_module.graph_adapter(original_graph)

        # Write to store type
        for store in stores:
            backend_module.write(original_graph, store, memory_geff["metadata"])
            new_graph = backend_module.graph_adapter(backend_module.read(store)[0])
            assert adpt_og_graph.get_node_ids() == new_graph.get_node_ids()
            assert adpt_og_graph.get_edge_ids() == new_graph.get_edge_ids()

    def test_overwrite(self, backend):
        backend_module: Backend = get_backend(backend)
        store_2d, _ = create_simple_2d_geff()
        _, memory_geff_3d = create_simple_3d_geff()

        graph = backend_module.construct(**memory_geff_3d)
        graph_adapter = backend_module.graph_adapter(graph)

        # Fails without overwrite
        with pytest.raises(FileExistsError, match="Found an existing geff present in `store`"):
            write(graph, store_2d)

        with pytest.raises(
            UserWarning,
            match="Cannot delete root zarr directory, but geff contents have been deleted",
        ):
            write(graph, store_2d, overwrite=True)
            _assert_graph_equal_to_geff(graph_adapter, memory_geff_3d)


@pytest.mark.parametrize("backend", get_args(SupportedBackend))
class Test_empty_graph:
    def test_read(self, backend):
        backend_module: Backend = get_backend(backend)

        store, memory_geff = create_empty_geff()

        graph, _metadata = read(store, backend=backend)
        graph_adapter = backend_module.graph_adapter(graph)

        _assert_graph_equal_to_geff(graph_adapter, memory_geff)

    def test_construct(self, backend):
        backend_module: Backend = get_backend(backend)

        store, memory_geff = create_empty_geff()

        in_memory_geff = read_to_memory(store)
        graph = construct(**in_memory_geff, backend=backend)
        graph_adapter = backend_module.graph_adapter(graph)

        _assert_graph_equal_to_geff(graph_adapter, memory_geff)

    def test_write(self, tmp_path, backend):
        if backend == "networkx":
            original_graph = nx.Graph()
        elif backend == "rustworkx":
            original_graph = rx.PyGraph()
        elif backend == "spatial-graph":
            create_graph = getattr(sg, "create_graph", sg.SpatialGraph)
            original_graph = create_graph(
                ndims=1,
                node_dtype="uint64",
                node_attr_dtypes={"pos": "float32[1]"},
                edge_attr_dtypes={},
                position_attr="pos",
            )
        else:
            raise NotImplementedError(
                f"Backend {backend} not tested in Test_empty_graph.test_write"
            )

        # write with unified write function
        path_store = tmp_path / "test_path.zarr"
        write(original_graph, path_store)

        # check that graph is empty after loading to in memory geff
        mem_geff = read_to_memory(path_store)
        assert len(mem_geff["node_ids"]) == 0
        assert len(mem_geff["edge_ids"]) == 0
        assert mem_geff["node_props"] == {}
        assert mem_geff["edge_props"] == {}



from __future__ import annotations

from typing import TYPE_CHECKING

import numpy as np

if TYPE_CHECKING:
    from numpy.typing import ArrayLike


def validate_unique_node_ids(node_ids: ArrayLike) -> tuple[bool, np.ndarray]:
    """Validates that all node ids are unique

    Args:
        node_ids (ArrayLike): 1D arraylike of node ids

    Returns:
        tuple[bool, np.ndarray]:
            - valid (bool): True if all node ids are unique
            - errors (list[ints]): List of any non-unique node ids
    """
    node_ids = np.asarray(node_ids)

    unique_ids, counts = np.unique(node_ids, return_counts=True)
    if any(counts > 1):
        return False, unique_ids[counts > 1]
    return True, np.array([])


def validate_nodes_for_edges(node_ids: ArrayLike, edge_ids: ArrayLike) -> tuple[bool, np.ndarray]:
    """
    Validates that all edges in `edge_ids` reference node IDs present in `node_ids`.

    This function checks whether each edge in `edge_ids` consists of node IDs that exist in
    `node_ids`. It returns a boolean indicating whether all edges are valid, and a list of
    invalid edges.

    Args:
        node_ids (ArrayLike): 1D array-like of valid node IDs (integers).
        edge_ids (ArrayLike): 2D array-like of edges with shape (M, 2), where each row is
            (source, target).

    Returns:
        tuple[bool, np.ndarray]:
            - all_edges_valid (bool): True if all edges reference valid node IDs.
            - invalid_edges (np.ndarray): Array of (source, target) pairs for
              invalid edges.
    """

    node_ids = np.asarray(node_ids)
    edge_ids = np.asarray(edge_ids)

    # Build a boolean mask: True for valid edges
    valid_src = np.isin(edge_ids[:, 0], node_ids)
    valid_tgt = np.isin(edge_ids[:, 1], node_ids)
    mask = valid_src & valid_tgt

    # Find invalid edges
    invalid_edges = edge_ids[~mask]
    all_edges_valid = len(invalid_edges) == 0
    return all_edges_valid, invalid_edges


def validate_no_self_edges(edge_ids: ArrayLike) -> tuple[bool, np.ndarray]:
    """
    Validates that there are no self-edges in the provided array of edges.

    Args:
        edge_ids (ArrayLike): 2D array-like of edges with shape (M, 2). Each row is
            (source, target).

    Returns:
        tuple[bool, np.ndarray]: A tuple (is_valid, problematic_nodes) where:
            - is_valid (bool): True if no node has an edge to itself, False otherwise.
            - problematic_nodes (np.ndarray): Array of node IDs that have self-edges.
              Empty if valid.
    """
    edge_ids = np.asarray(edge_ids)
    mask = edge_ids[:, 0] == edge_ids[:, 1]
    problematic_nodes = np.unique(edge_ids[mask, 0])
    return (len(problematic_nodes) == 0, problematic_nodes)


def validate_no_repeated_edges(edge_ids: ArrayLike) -> tuple[bool, np.ndarray]:
    """
    Validates that there are no repeated edges in the array.

    Args:
        edge_ids (ArrayLike): 2D array-like of edges with shape (M, 2). Each row is
            (source, target).

    Returns:
        tuple: A tuple (is_valid, repeated_edges) where:
            - is_valid (bool): True if there are no repeated edges, False otherwise.
            - repeated_edges (np.ndarray): An array of duplicated edges. Empty if valid.

    """
    edge_ids = np.asarray(edge_ids)
    edges_view = np.ascontiguousarray(edge_ids).view([("", edge_ids.dtype)] * edge_ids.shape[1])
    _, idx, counts = np.unique(edges_view, return_index=True, return_counts=True)
    repeated_mask = counts > 1
    repeated_edges = edge_ids[idx[repeated_mask]]
    return (len(repeated_edges) == 0, repeated_edges)



from __future__ import annotations

import itertools
import warnings
from copy import deepcopy
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import networkx as nx

from geff.core_io._utils import check_for_geff, delete_geff

if TYPE_CHECKING:
    import xml.etree.ElementTree as ET
    from collections.abc import Callable, Container, Iterator

    # from lxml import etree as ET
else:
    # Prefer lxml for performance, but gracefully fall back to the Python
    # standard-library implementation to avoid the hard dependency.
    # these follow a similar enough API that we can use the same code.
    try:
        from lxml import etree as ET
    except ImportError:  # pragma: no cover
        import xml.etree.ElementTree as ET

from geff._graph_libs._networkx import NxBackend
from geff_spec import Axis, DisplayHint, GeffMetadata, RelatedObject

# Template mapping for TrackMate dimension to unit conversion
_DIMENSION_UNIT_TEMPLATES: dict[str, Callable[[str, str], str]] = {
    "NONE": lambda space, time: "None",
    "QUALITY": lambda space, time: "None",
    "COST": lambda space, time: "None",
    "INTENSITY": lambda space, time: "None",
    "INTENSITY_SQUARED": lambda space, time: "None",
    "STRING": lambda space, time: "None",
    "POSITION": lambda space, time: space,
    "LENGTH": lambda space, time: space,
    "TIME": lambda space, time: time,
    "VELOCITY": lambda space, time: f"{space} / {time}",
    "AREA": lambda space, time: f"{space}^2",
    "ANGLE": lambda space, time: "radian",
    "RATE": lambda space, time: f"1 / {time}",
    "ANGLE_RATE": lambda space, time: f"radian / {time}",
}


def _preliminary_checks(
    xml_path: Path, geff_path: Path, overwrite: bool, zarr_format: Literal[2, 3] = 2
) -> None:
    """Check the validity of input paths and clean up the output path if needed.

    Args:
        xml_path (Path): The path to the TrackMate XML file.
        geff_path (Path): The path to the GEFF file.
        overwrite (bool): Whether to overwrite the GEFF file if it already exists.
        zarr_format (Literal[2, 3], optional): The version of zarr to write. Defaults to 2.

    Raises:
        FileNotFoundError: If the XML file does not exist.
        FileExistsError: If the GEFF file exists and overwrite is False.
    """
    if not xml_path.exists():
        raise FileNotFoundError(f"TrackMate XML file {xml_path} does not exist.")

    # Check for an existing geff
    if check_for_geff(geff_path):
        if overwrite:
            delete_geff(geff_path, zarr_format=zarr_format)
        else:
            raise FileExistsError(
                "Found an existing geff present in `geff_store`. "
                "Please use `overwrite=True` or provide an alternative "
                "`geff_store` to write to."
            )


def _get_units(
    element: ET.Element,
) -> dict[str, str]:
    """Extract units information from an XML element and return it as a dictionary.

    This function deep copies the attributes of the XML element into a dictionary,
    then clears the element to free up memory.

    Args:
        element (ET._Element): The XML element holding the units information.

    Returns:
        dict[str, str]: A dictionary containing the units information.
        Keys are 'spatialunits' and 'timeunits'.

    Warns:
        UserWarning: If the 'spatialunits' or 'timeunits' attributes are not found,
            defaulting them to 'pixel' and 'frame', respectively.
    """
    units = {}  # type: dict[str, str]
    if element.attrib:
        units = deepcopy(element.attrib)
    if "spatialunits" not in units:
        warnings.warn(
            "No space unit found in the XML file. Setting to 'pixel'.",
            stacklevel=4,
        )
        units["spatialunits"] = "pixel"  # TrackMate default value.
    if "timeunits" not in units:
        warnings.warn(
            "No time unit found in the XML file. Setting to 'frame'.",
            stacklevel=4,
        )
        units["timeunits"] = "frame"  # TrackMate default value.
    element.clear()  # We won't need it anymore so we free up some memory.
    # .clear() does not delete the element: it only removes all subelements
    # and clears or sets to `None` all attributes.
    return units


def _get_attributes_metadata(
    it: Iterator[tuple[str, ET.Element]],
    ancestor: ET.Element,
) -> dict[str, dict[str, str]]:
    """Extract the TrackMate model features to a attributes dictionary.

    The model features are divided in 3 categories: SpotFeatures, EdgeFeatures and
    TrackFeatures. Those features are regrouped under the FeatureDeclarations tag.
    Some other features are used in the Spot and Track tags but are not declared in
    the FeatureDeclarations tag.

    Args:
        it (Iterator[tuple[str, ET.Element]]): An iterator over XML elements.
        ancestor (ET._Element): The XML element that encompasses the information to be added.

    Returns:
        dict[str, dict[str, str]]: A dictionary where the keys are the attributes names
        and the values are dictionaries containing the attributes metadata as defined by TrackMate
        (name, shortname, dimension, isint).
    """
    attrs_md = {}
    event, element = next(it)
    while (event, element) != ("end", ancestor):
        # Features stored in the FeatureDeclarations tag.
        event, element = next(it)  # Feature.
        while (event, element) != ("end", ancestor):
            if element.tag == "Feature" and event == "start":
                attrs = deepcopy(element.attrib)
                attrs_md[attrs["feature"]] = attrs
                attrs_md[attrs["feature"]].pop("feature", None)
            element.clear()
            event, element = next(it)
    return attrs_md


def _convert_attributes(
    attrs: dict[str, Any],
    attrs_metadata: dict[str, dict[str, str]],
    attr_type: str,
    stack_level: int,
) -> None:
    """
    Convert the values of the attributes from string to the correct data type.

    TrackMate features are either integers, floats or strings. The type to
    convert to is given by the attributes metadata.

    Args:
        attrs (dict[str, Any): The dictionary whose values we want to convert.
        attrs_metadata (dict[str, dict[str, str]]): The attributes metadata containing
        information on the expected data types for each attribute.
        attr_type (str): The type of the attribute to convert (node, edge, or lineage).
        stack_level (int): The stack level for warnings, used to indicate where
            the warning originated.

    Raises:
        ValueError: If an attribute value cannot be converted to the expected type.

    Warns:
        UserWarning: If an attribute is not found in the attributes metadata.
    """
    for key in attrs:
        if key in attrs_metadata:
            if attrs_metadata[key]["isint"] == "true":
                try:
                    attrs[key] = int(attrs[key])
                except ValueError as err:
                    raise ValueError(f"Invalid integer value for {key}: {attrs[key]}") from err
            else:
                try:
                    attrs[key] = float(attrs[key])
                except ValueError:
                    # Then it's a string and no need to convert.
                    pass
        elif key == "ID" or key == "ROI_N_POINTS":
            # IDs are always integers in TrackMate.
            attrs[key] = int(attrs[key])
        elif key == "name":
            pass  # "name" is a string so we don't need to convert it.
        else:
            warnings.warn(
                f"{attr_type.capitalize()} attribute {key} not found in the attributes metadata.",
                stacklevel=stack_level,
            )


def _convert_ROI_coordinates(
    element: ET.Element,
    attrs: dict[str, Any],
) -> int:
    """Extract, format and add ROI coordinates to the attributes dict.

    Args:
        element (ET._Element): Element from which to extract ROI coordinates.
        attrs (dict[str, Any]): Attributes dict to update with ROI coordinates.

    Returns:
        int: Number of dimensions of the ROI coordinates.

    Raises:
        KeyError: If the "ROI_N_POINTS" attribute is not found in the attributes dict.
        TypeError: If the "ROI_N_POINTS" attribute is not an integer.
    """
    if "ROI_N_POINTS" not in attrs:
        raise KeyError(
            f"No key 'ROI_N_POINTS' in the attributes of current element '{element.tag}'."
        )
    if element.text:
        n_points = attrs["ROI_N_POINTS"]
        if not isinstance(n_points, int):
            raise TypeError("ROI_N_POINTS should be an integer.")

        coords = [float(v) for v in element.text.split()]
        nb_dim = len(coords) // n_points
        attrs["ROI_coords"] = [tuple(coords[i : i + nb_dim]) for i in range(0, len(coords), nb_dim)]
    else:
        attrs["ROI_coords"] = None
        nb_dim = 0

    return nb_dim


def _add_all_nodes(
    it: Iterator[tuple[str, ET.Element]],
    ancestor: ET.Element,
    attrs_md: dict[str, dict[str, str]],
    graph: nx.DiGraph,
) -> bool:
    """Add nodes and their attributes to a graph and return the presence of segmentation.

    All the elements that are descendants of `ancestor` are explored. The graph is
    modified in place by adding nodes with their attributes.

    Args:
        it (Iterator[tuple[str, ET.Element]]): An iterator over XML elements.
        ancestor (ET._Element): The XML element that encompasses the information to be added.
        attrs_md (dict[str, dict[str, str]]): The attributes metadata containing the
            expected node attributes.
        graph (nx.DiGraph): The graph to which the nodes will be added (modified in place).

    Returns:
        bool: True if the spots are segmented, False otherwise.

    Warns:
        UserWarning: If a node cannot be added to the graph due to missing attributes.
    """
    segmentation = False
    event, element = next(it)
    while (event, element) != ("end", ancestor):
        event, element = next(it)
        if element.tag == "Spot" and event == "end":
            # All items in element.attrib are parsed as strings but most
            # of them are numbers. So we need to do a conversion based
            # on these attributes type as defined in the attributes
            # metadata (attribute `isint`).
            attrs = deepcopy(element.attrib)
            _convert_attributes(attrs, attrs_md, "node", 5)

            # The ROI coordinates are not stored in a tag attribute but in
            # the tag text. So we need to extract then format them.
            # In case of a single-point detection, the `ROI_N_POINTS` attribute
            # is not present.
            if segmentation:
                _convert_ROI_coordinates(element, attrs)
            else:
                if "ROI_N_POINTS" in attrs:
                    segmentation = True
                    _convert_ROI_coordinates(element, attrs)

            # Adding the node and its attributes to the graph.
            try:
                graph.add_node(attrs["ID"], **attrs)
            except KeyError:
                warnings.warn(
                    f"No key 'ID' in the attributes of current element "
                    f"'{element.tag}'. Not adding this node to the graph.",
                    stacklevel=4,
                )
            finally:
                element.clear()

    return segmentation


def _add_edge(
    element: ET.Element,
    attrs_md: dict[str, dict[str, str]],
    graph: nx.DiGraph,
    current_track_id: int,
) -> None:
    """Add an edge between two nodes in the graph based on the XML element.

    This function extracts source and target node identifiers from the
    given XML element, along with any additional attributes defined
    within. It then adds an edge between these nodes in the specified
    graph. If the nodes have a 'TRACK_ID' attribute, it ensures consistency
    with the current track ID. The graph is modified in place.

    Args:
        element (ET._Element): The XML element containing edge information.
        attrs_md (dict[str, dict[str, str]]): The attributes metadata containing
            the expected edge attributes.
        graph (nx.DiGraph): The graph to which the edge and its attributes will be added
            (modified in place).
        current_track_id (int): Track ID of the track holding the edge.

    Raises:
        AssertionError: If the 'TRACK_ID' attribute of either the source or target node
            does not match the current track ID, indicating an inconsistency in track
            assignment.

    Warns:
        UserWarning: If an edge cannot be added due to missing required attributes.
    """
    attrs = deepcopy(element.attrib)
    _convert_attributes(attrs, attrs_md, "edge", 6)
    try:
        entry_node_id = int(attrs["SPOT_SOURCE_ID"])
        exit_node_id = int(attrs["SPOT_TARGET_ID"])
    except KeyError:
        warnings.warn(
            f"No key 'SPOT_SOURCE_ID' or 'SPOT_TARGET_ID' in the attributes of "
            f"current element '{element.tag}'. Not adding this edge to the graph.",
            stacklevel=5,
        )
    else:
        graph.add_edge(entry_node_id, exit_node_id)
        nx.set_edge_attributes(graph, {(entry_node_id, exit_node_id): attrs})
        # Adding the current track ID to the nodes of the newly created
        # edge. This will be useful later to filter nodes by track and
        # add the saved tracks attributes (as returned by this method).
        err_msg = f"Incoherent track ID for nodes {entry_node_id} and {exit_node_id}."
        entry_node = graph.nodes[entry_node_id]
        if "TRACK_ID" not in entry_node:
            entry_node["TRACK_ID"] = current_track_id
        else:
            assert entry_node["TRACK_ID"] == current_track_id, err_msg
        exit_node = graph.nodes[exit_node_id]
        if "TRACK_ID" not in exit_node:
            exit_node["TRACK_ID"] = current_track_id
        else:
            assert exit_node["TRACK_ID"] == current_track_id, err_msg
    finally:
        element.clear()


def _build_tracks(
    iterator: Iterator[tuple[str, ET.Element]],
    ancestor: ET.Element,
    attrs_md: dict[str, dict[str, str]],
    graph: nx.DiGraph,
) -> list[dict[str, Any]]:
    """Add edges and their attributes to a graph based on the XML elements.

    This function explores all elements that are descendants of the
    specified `ancestor` element, adding edges and their attributes to
    the provided graph. It iterates through the XML elements using
    the provided iterator, extracting and processing relevant information
    to construct track attributes. The graph is modified in place.

    Args:
        iterator (Iterator[tuple[str, ET.Element]]): An iterator over XML elements.
        ancestor (ET._Element): The XML element that encompasses the information to be added.
        attrs_md (dict[str, dict[str, str]]): The attributes metadata containing the
            expected edge attributes.
        graph (nx.DiGraph): The graph to which the edges and their attributes will be added
            (modified in place).

    Returns:
        list[dict[str, Any]]: A list of dictionaries, each representing the
            attributes for a track.

    Raises:
        KeyError: If no TRACK_ID is found in the attributes of a Track element.
    """
    tracks_attrs = []
    current_track_id = None
    event, element = next(iterator)
    while (event, element) != ("end", ancestor):
        # Saving the current track information.
        if element.tag == "Track" and event == "start":
            attrs: dict[str, Any] = deepcopy(element.attrib)
            _convert_attributes(attrs, attrs_md, "lineage", 5)
            tracks_attrs.append(attrs)
            try:
                current_track_id = attrs["TRACK_ID"]
            except KeyError as err:
                raise KeyError(
                    f"No key 'TRACK_ID' in the attributes of current element "
                    f"'{element.tag}'. Please check the XML file.",
                ) from err

        # Edge creation.
        if element.tag == "Edge" and event == "start":
            assert current_track_id is not None, "No current track ID."
            _add_edge(element, attrs_md, graph, current_track_id)

        event, element = next(iterator)

    return tracks_attrs


def _get_filtered_tracks_ID(
    iterator: Iterator[tuple[str, ET.Element]],
    ancestor: ET.Element,
) -> list[int]:
    """
    Extract and return a list of track IDs identifying the tracks to keep.

    This function processes the first element immediately, then iterates through
    the remaining elements to extract track IDs from TrackID elements.

    Args:
        iterator (Iterator[tuple[str, ET.Element]]): An iterator over XML elements.
        ancestor (ET._Element): The XML element that encompasses the information to be added.

    Returns:
        list[int]: List of tracks ID to identify the tracks to keep.

    Warns:
        UserWarning: If the "TRACK_ID" attribute is not found in the attributes.
    """
    filtered_tracks_ID = []
    event, element = next(iterator)
    attrs = deepcopy(element.attrib)
    try:
        filtered_tracks_ID.append(int(attrs["TRACK_ID"]))
    except KeyError:
        warnings.warn(
            f"No key 'TRACK_ID' in the attributes of current element "
            f"'{element.tag}'. Ignoring this track.",
            stacklevel=4,
        )

    while (event, element) != ("end", ancestor):
        event, element = next(iterator)
        if element.tag == "TrackID" and event == "start":
            attrs = deepcopy(element.attrib)
            try:
                filtered_tracks_ID.append(int(attrs["TRACK_ID"]))
            except KeyError:
                warnings.warn(
                    f"No key 'TRACK_ID' in the attributes of current element "
                    f"'{element.tag}'. Ignoring this track.",
                    stacklevel=4,
                )

    return filtered_tracks_ID


def _build_data(
    xml_path: Path,
    discard_filtered_spots: bool = False,
    discard_filtered_tracks: bool = False,
) -> tuple[nx.DiGraph, dict[str, str], bool]:
    """Read an XML file and convert the model data into several graphs.

    All TrackMate tracks and their associated data described in the XML file
    are modeled as a networkX graph. Spots are modeled as graph
    nodes, and edges as graph edges. Spot, edge and track features are
    stored in node, edge and graph attributes, respectively.

    Args:
        xml_path (Path): Path of the XML file to process.
        discard_filtered_spots (bool, optional): True to discard the spots
            filtered out in TrackMate, False otherwise. False by default.
        discard_filtered_tracks (bool, optional): True to discard the tracks
            filtered out in TrackMate, False otherwise. False by default.

    Returns:
        nx.DiGraph: A NetworkX graph representing the TrackMate data.
        dict[str, str]: A dictionary containing the units of the model, with keys
            'spatialunits' and 'timeunits'.
        bool: True if the spots are segmented, False otherwise.
    """
    graph: nx.DiGraph[int] = nx.DiGraph()
    segmentation = False

    # So as not to load the entire XML file into memory at once, we're
    # using an iterator to browse over the tags one by one.
    # The events 'start' and 'end' correspond respectively to the opening
    # and the closing of the considered tag.
    with open(xml_path, "rb") as f:
        it = ET.iterparse(f, events=["start", "end"])
        _, root = next(it)  # Saving the root of the tree for later cleaning.

        units: dict[str, str] = {}
        attrs_md: dict[str, dict[str, str]] = {}
        for event, element in it:
            if element.tag == "Model" and event == "start":
                units = _get_units(element)
                root.clear()  # Cleaning the tree to free up some memory.
                # All the browsed subelements of `root` are deleted.

            # Get the spot, edge and track features and add them to the
            # attributes metadata.
            if element.tag == "FeatureDeclarations" and event == "start":
                attrs_md = _get_attributes_metadata(it, element)
                root.clear()

            # Adding the spots as nodes.
            if element.tag == "AllSpots" and event == "start":
                segmentation = _add_all_nodes(it, element, attrs_md, graph)
                root.clear()

            # Adding the tracks as edges.
            if element.tag == "AllTracks" and event == "start":
                # TODO: implement storage of track attributes.
                # tracks_attrs = _build_tracks(it, element, attrs_md, graph)
                _build_tracks(it, element, attrs_md, graph)
                root.clear()

                # Removal of filtered spots / nodes.
                if discard_filtered_spots:
                    # Those nodes belong to no tracks: they have a degree of 0.
                    # TODO: remove ignore, see issue 314
                    lone_nodes = [n for n, d in graph.degree if d == 0]  # pyright: ignore[reportGeneralTypeIssues]
                    graph.remove_nodes_from(lone_nodes)

            # Filtering out tracks.
            if element.tag == "FilteredTracks" and event == "start":
                # Removal of filtered tracks.
                id_to_keep: Container = _get_filtered_tracks_ID(it, element)
                if discard_filtered_tracks:
                    to_remove = [
                        n
                        # TODO: remove ignore, see issue 314
                        for n, t in graph.nodes(data="TRACK_ID")  # pyright: ignore[reportGeneralTypeIssues]
                        if t is None or t not in id_to_keep
                    ]
                    graph.remove_nodes_from(to_remove)

            if element.tag == "Model" and event == "end":
                break

    return graph, units, segmentation


def _get_trackmate_version(
    xml_path: Path,
) -> str:
    """
    Extract the version of TrackMate used to generate the XML file.

    Args:
    xml_path (Path): The file path of the XML file to be parsed.

    Returns:
        str: The version of TrackMate used to generate the XML file.
            If the version cannot be found, "unknown" is returned.
    """
    with open(xml_path, "rb") as f:
        it = ET.iterparse(f)
        for _, element in it:
            if element.tag == "TrackMate":
                version = element.attrib.get("version")
                return str(version) if version else "unknown"
            else:
                element.clear()
    return "unknown"


def _get_specific_tags(
    xml_path: Path,
    tag_names: list[str],
    stack_level: int,
) -> dict[str, ET.Element]:
    """
    Extract specific tags from an XML file and return them in a dictionary.

    This function parses an XML file, searching for specific tag names.
    Once a tag is found, the children elements are extracted and stored
    in a dictionary with the tag name as the key. The search stops when
    all specified tags have been found or when the end of the file is reached.

    Args:
    xml_path (Path): The file path of the XML file to be parsed.
    tag_names (list[str]): A list of tag names to search for in the XML file.
    stack_level (int): The stack level for warnings, used to indicate where
        the warning originated.

    Returns:
        dict[str, ET.Element]: A dictionary where each key is a tag name from
        `tag_names` that was found in the XML file, and the corresponding value
        is the deep copied `ET.Element` object for that tag.

    Warns:
        UserWarning: If any of the specified tags in `tag_names` are not found
            in the XML file. The error message will list the missing tags.
    """
    with open(xml_path, "rb") as f:
        it = ET.iterparse(f, events=["start", "end"])
        dict_tags = {}
        for event, element in it:
            if event == "start" and element.tag in tag_names:
                dict_tags[element.tag] = deepcopy(element)
                tag_names.remove(element.tag)
                if not tag_names:  # All the tags have been found.
                    break

            if event == "end":
                element.clear()

        if tag_names:
            warnings.warn(
                f"Missing tag(s): {tag_names}. The associated metadata "
                "will not be included in the GEFF file.",
                stacklevel=stack_level,
            )

    return dict_tags


def _extract_image_path(settings_md: ET.Element | None) -> str | None:
    """Extract the image path from the TrackMate settings metadata.

    Args:
        settings_md (ET.Element | None): The XML element containing the settings metadata.

    Returns:
        str | None: The image path if found, otherwise None.

    Warnings:
        UserWarning: If the 'Settings' or 'ImageData' tags are not found in the XML file,
            or if the image path cannot be constructed from the available data.
    """
    if settings_md is None:
        warnings.warn(
            (
                "No 'Settings' tag found in the XML file. "
                "The GEFF file will not point to a related image."
            ),
            stacklevel=3,
        )
        return None
    image_data = settings_md.find("ImageData")
    if image_data is None:
        warnings.warn(
            (
                "No 'ImageData' tag found in the XML file. "
                "The GEFF file will not point to a related image."
            ),
            stacklevel=3,
        )
        return None

    filename = image_data.attrib["filename"]
    folder = image_data.attrib["folder"]
    if not filename and not folder:
        warnings.warn(
            "No image path found in the XML file. The GEFF file will not point to a related image.",
            stacklevel=3,
        )
        return None
    elif not filename:
        warnings.warn(
            "No image name found in the XML file. The GEFF file will only point to a folder.",
            stacklevel=3,
        )
        return str(Path(folder))
    elif not folder:
        warnings.warn(
            "No image folder found in the XML file. "
            "The GEFF file will only point to an image name.",
            stacklevel=3,
        )
        return str(Path(filename))
    else:
        return str(Path(folder) / filename)


def _get_feature_name(feat: ET.Element, key: str, feat_type: str) -> str:
    """Extract and validate the feature name, using key as fallback.

    Args:
        feat: The feature metadata element.
        key: The feature identifier to use as fallback.
        feat_type: The feature type for warning messages.

    Returns:
        The feature name.
    """
    name = feat.attrib.get("name", None)
    if name is None:
        warnings.warn(
            f"Missing field 'name' in 'FeatureDeclarations' {feat_type} tag. "
            "Using the feature identifier as name.",
            stacklevel=5,
        )
        name = key
    return name


def _get_feature_dtype(feat: ET.Element, feat_type: str) -> str:
    """Extract and convert the feature data type.

    Args:
        feat: The feature metadata element.
        feat_type: The feature type for error messages.

    Returns:
        The feature data type ('int' or 'float').

    Raises:
        ValueError: If the dtype field is missing.
    """
    dtype = feat.attrib.get("isint", None)
    if dtype is None:
        raise ValueError(
            f"Missing field 'isint' in 'FeatureDeclarations' {feat_type} tag. "
            "Please check the XML file."
        )
    return "int" if dtype == "true" else "float"


def _get_feature_unit(feat: ET.Element, feat_type: str, units: dict[str, str]) -> str:
    """Infer the feature unit from TrackMate 'dimension'.

    Args:
        feat: The feature metadata element.
        feat_type: The feature type for warning messages.
        units: A dictionary containing the units of the model.

    Returns:
        The feature unit.
    """
    dimension = feat.attrib.get("dimension", None)
    if dimension not in _DIMENSION_UNIT_TEMPLATES:
        raise ValueError(
            f"Unknown dimension '{dimension}' in 'FeatureDeclarations' '{feat_type}' tag. "
            "Please check the XML file."
        )
    space = units.get("spatialunits", "pixel")
    time = units.get("timeunits", "frame")

    return _DIMENSION_UNIT_TEMPLATES[dimension](space, time)


def _process_feature_metadata(
    feat: ET.Element, geff_dict: dict[str, Any], feat_type: str, units: dict[str, str]
) -> None:
    """Process a single feature metadata entry and add it to props_metadata.

    Args:
        feat: The feature metadata element from XML.
        geff_dict: The dictionary to update with the processed metadata.
        feat_type: The feature type (for error/warning messages).
        units: A dictionary containing the units of the model.

    Raises:
        KeyError: If the 'feature' key is missing from the XML element.
        ValueError: If the 'feature' key is duplicated.
    """
    if feat.attrib.get("feature") is None:
        raise KeyError(
            f"Missing field 'feature' in 'FeatureDeclarations' {feat_type} tag. "
            "Please check the XML file."
        )
    else:
        key = feat.attrib["feature"]

    if key in geff_dict:
        raise ValueError(
            f"Duplicate feature identifier '{key}' found in 'FeatureDeclarations' tag."
        )

    geff_dict[key] = {
        "identifier": key,
        "name": _get_feature_name(feat, key, feat_type),
        "dtype": _get_feature_dtype(feat, feat_type),
        "unit": _get_feature_unit(feat, feat_type, units),
    }


def _extract_props_metadata(
    xml_path: Path, units: dict[str, str], segmentation: bool
) -> dict[str, Any]:
    """Extract properties metadata from the TrackMate XML file.

    Args:
        xml_path: The file path of the XML file to be parsed.
        units: A dictionary containing the units of the model.
        segmentation: A boolean indicating whether the spots are segmented.

    Returns:
        dict[str, Any]
            A dictionary containing the properties metadata extracted from the XML file.

    Raises:
        ValueError: If the 'FeatureDeclarations' tag is not found in the XML file,
            if a required field is missing in the 'FeatureDeclarations' tag,
            or if a duplicate feature identifier is found.

    Warnings:
        UserWarning: If a feature identifier is missing a 'name' or 'shortname'
            field in the 'FeatureDeclarations' tag, a default value will be used.
            If a feature identifier is missing a 'dimension' field in the 'FeatureDeclarations' tag,
            None will be used.
    """
    # Dictionaries of metadata for GEFF properties.
    node_props_metadata: dict[str, Any] = {}
    edge_props_metadata: dict[str, Any] = {}
    lineage_props_metadata: dict[str, Any] = {}
    props_metadata = {
        "node_props_metadata": node_props_metadata,
        "edge_props_metadata": edge_props_metadata,
        "lineage_props_metadata": lineage_props_metadata,
    }
    # Mapping TrackMate feature types to GEFF metadata fields.
    mapping_feat_type = {
        "SpotFeatures": node_props_metadata,
        "EdgeFeatures": edge_props_metadata,
        "TrackFeatures": lineage_props_metadata,
    }

    tags_data = _get_specific_tags(xml_path, ["FeatureDeclarations"], 4)
    xml_md = tags_data["FeatureDeclarations"]
    for feat_type in xml_md:
        if feat_type.tag in mapping_feat_type:
            for feat in feat_type.findall("Feature"):
                _process_feature_metadata(
                    feat, mapping_feat_type[feat_type.tag], feat_type.tag, units
                )

    # Specific case of ROI.
    if segmentation:
        node_props_metadata["ROI_N_POINTS"] = {
            "identifier": "ROI_N_POINTS",
            "dtype": "int",
            "name": "ROI number of points",
            "description": "Number of points defining the spot ROI.",
        }
        node_props_metadata["ROI_coords"] = {
            "identifier": "ROI_coords",
            "dtype": "float",
            "varlength": True,
            "unit": node_props_metadata["POSITION_X"].get("unit", "pixel"),
            "name": "ROI coordinates",
            "description": "List of coordinates of the spot ROI relative to the spot center.",
        }

    return props_metadata


def _build_geff_metadata(
    xml_path: Path,
    units: dict[str, str],
    img_path: str | None,
    trackmate_metadata: dict[str, ET.Element],
    props_metadata: dict[str, dict[str, Any]],
) -> GeffMetadata:
    """Create GEFF metadata from TrackMate XML data.

    Args:
        xml_path (Path): The path to the TrackMate XML file.
        units (dict[str, str]): A dictionary containing the units of the model.
        img_path (str | None): The path to the related image file.
        trackmate_metadata (dict[str, ET.Element]): The TrackMate metadata extracted
            from the XML file.
        props_metadata (dict[str, Any]): The properties metadata extracted from the XML file.

    Returns:
        GeffMetadata: The constructed GEFF metadata object.
    """
    extra = {
        "other_trackmate_metadata": {
            "trackmate_version": _get_trackmate_version(xml_path),
            # TODO: move into normal metadata once GEFF can store lineage metadata
            "lineage_props_metadata": props_metadata["lineage_props_metadata"],
        },
    }
    md = extra["other_trackmate_metadata"]
    if "Log" in trackmate_metadata:
        md["log"] = ET.tostring(trackmate_metadata["Log"], encoding="utf-8").decode()
    if "Settings" in trackmate_metadata:
        md["settings"] = ET.tostring(trackmate_metadata["Settings"], encoding="utf-8").decode()
    if "GUIState" in trackmate_metadata:
        md["gui_state"] = ET.tostring(trackmate_metadata["GUIState"], encoding="utf-8").decode()
    if "DisplaySettings" in trackmate_metadata:
        md["display_settings"] = ET.tostring(
            trackmate_metadata["DisplaySettings"], encoding="utf-8"
        ).decode()

    return GeffMetadata(
        axes=[
            Axis(name="POSITION_X", type="space", unit=units.get("spatialunits", "pixel")),
            Axis(name="POSITION_Y", type="space", unit=units.get("spatialunits", "pixel")),
            Axis(name="POSITION_Z", type="space", unit=units.get("spatialunits", "pixel")),
            Axis(name="POSITION_T", type="time", unit=units.get("timeunits", "frame")),
        ],
        display_hints=DisplayHint(
            display_horizontal="POSITION_X",
            display_vertical="POSITION_Y",
            display_depth="POSITION_Z",
            display_time="POSITION_T",
        ),
        directed=True,
        node_props_metadata=props_metadata["node_props_metadata"],
        edge_props_metadata=props_metadata["edge_props_metadata"],
        track_node_props={"lineage": "TRACK_ID"},
        related_objects=[RelatedObject(type="image", path=img_path)] if img_path else None,
        extra=extra,
    )


def _check_component_props_consistency(
    component_data: Iterator[dict[str, Any]],
    metadata_dict: dict[str, Any],
    component_type: str,
) -> None:
    """Check consistency between component data and metadata properties.

    Args:
        component_data: Iterator over component data dictionaries.
        metadata_dict: Dictionary containing metadata for the component type.
        component_type: Type of component ("node", "edge", or "lineage").

    Note:
        If any properties defined in the metadata are not present in the graph data,
        they will be removed from the metadata and an info message will be logged.
    """
    props_to_check = list(metadata_dict.keys())
    iterators = itertools.tee(component_data, len(props_to_check))  # one iterator per prop
    removed_props = []
    for prop, data_iter in zip(props_to_check, iterators, strict=True):
        if not any(prop in data for data in data_iter):  # early termination when prop found
            metadata_dict.pop(prop)
            removed_props.append(prop)

    if removed_props:
        plural1 = "ies were" if len(removed_props) > 1 else "y was"
        plural2 = "they are" if len(removed_props) > 1 else "it is"
        warnings.warn(
            f"The following {component_type} propert{plural1} removed from the metadata "
            f"because {plural2} not present in the data: {', '.join(removed_props)}.",
            stacklevel=4,
        )


def _ensure_data_metadata_consistency(
    graph: nx.DiGraph,
    metadata: dict[str, Any],
) -> None:
    """Ensure that the graph data and metadata are consistent.

    Geff specification requires that all metadata properties
    defined in the GEFF file are present in the graph data.

    Args:
        graph (nx.DiGraph): The graph to check.
        metadata (dict[str, Any]): The metadata to check and update.
    """
    # Nodes
    _check_component_props_consistency(
        component_data=(node_data for _, node_data in graph.nodes(data=True)),
        metadata_dict=metadata.get("node_props_metadata", {}),
        component_type="node",
    )
    # Edges
    _check_component_props_consistency(
        component_data=(edge_data for _, _, edge_data in graph.edges(data=True)),
        metadata_dict=metadata.get("edge_props_metadata", {}),
        component_type="edge",
    )


def from_trackmate_xml_to_geff(
    xml_path: Path | str,
    geff_path: Path | str,
    discard_filtered_spots: bool = False,
    discard_filtered_tracks: bool = False,
    overwrite: bool = False,
    zarr_format: Literal[2, 3] = 2,
) -> None:
    """
    Convert a TrackMate XML file to a GEFF file.

    Args:
        xml_path (Path | str): The path to the TrackMate XML file.
        geff_path (Path | str): The path to the GEFF file.
        discard_filtered_spots (bool, optional): True to discard the spots
            filtered out in TrackMate, False otherwise. False by default.
        discard_filtered_tracks (bool, optional): True to discard the tracks
            filtered out in TrackMate, False otherwise. False by default.
        overwrite (bool, optional): Whether to overwrite the GEFF file if it already exists.
        zarr_format (Literal[2, 3], optional): The version of zarr to write. Defaults to 2.

    Warns:
        UserWarning: If the XML file does not contain specific metadata tags or if there are issues
            with the TrackMate metadata.
    """
    xml_path = Path(xml_path)
    geff_path = Path(geff_path).with_suffix(".geff")
    _preliminary_checks(xml_path, geff_path, overwrite=overwrite, zarr_format=zarr_format)

    # Data
    graph, units, segmentation = _build_data(
        xml_path=xml_path,
        discard_filtered_spots=discard_filtered_spots,
        discard_filtered_tracks=discard_filtered_tracks,
    )
    # Metadata
    props_metadata = _extract_props_metadata(xml_path, units, segmentation)
    _ensure_data_metadata_consistency(graph=graph, metadata=props_metadata)
    tm_md = _get_specific_tags(xml_path, ["Log", "Settings", "GUIState", "DisplaySettings"], 3)
    img_path = _extract_image_path(tm_md.get("Settings", None))
    metadata = _build_geff_metadata(
        xml_path=xml_path,
        units=units,
        img_path=img_path,
        trackmate_metadata=tm_md,
        props_metadata=props_metadata,
    )

    # Create the GEFF :D
    NxBackend.write(
        graph,
        store=geff_path,
        metadata=metadata,
        zarr_format=zarr_format,
    )



from importlib.metadata import PackageNotFoundError, version

try:
    __version__: str = version("geff")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "uninstalled"

from geff_spec import GeffMetadata

from ._graph_libs._api_wrapper import SupportedBackend, construct, read, write
from .core_io._base_read import GeffReader
from .validate.structure import validate_structure

__all__ = [
    "GeffMetadata",
    "GeffReader",
    "SupportedBackend",
    "construct",
    "read",
    "validate_structure",
    "write",
]



"""Test data generation utilities for geff graphs.

This module provides functions to create mock geff graphs for testing and development.
It includes both simple convenience functions and a comprehensive function for advanced use cases.


Examples:
    # Simple 2D graph with defaults
    >>> store, memory_geff = create_simple_2d_geff()
    >>> # Creates: 10 nodes, 15 edges, undirected, 2D (x, y, t)

    # Simple 3D graph with custom size
    >>> store, memory_geff = create_simple_3d_geff(num_nodes=20, num_edges=30)
    >>> # Creates: 20 nodes, 30 edges, undirected, 3D (x, y, z, t)

    # Advanced usage with full control
    >>> store, memory_geff = create_mock_geff(
    ...     node_id_dtype="int",
    ...     node_axis_dtypes={"position": "float64", "time": "float32"},
    ...     directed=True,
    ...     num_nodes=5,
    ...     num_edges=8,
    ...     extra_node_props={"label": "str", "confidence": "float64"},
    ...     extra_edge_props={"score": "float64", "color": "uint8",
    ...           "weight": "float64", "type": "str"},
    ...     include_t=True,
    ...     include_z=False,  # 2D only
    ...     include_y=True,
    ...     include_x=True,
    ... )

    # Advanced usage with custom arrays
    >>> import numpy as np
    >>> custom_labels = np.array(["A", "B", "C", "D", "E"])
    >>> custom_scores = np.array([0.1, 0.5, 0.8, 0.3, 0.9])
    >>> custom_edge_weights = np.array([0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8])
    >>> store, memory_geff = create_mock_geff(
    ...     node_id_dtype="int",
    ...     node_axis_dtypes={"position": "float64", "time": "float64"},
    ...     directed=False,
    ...     num_nodes=5,
    ...     num_edges=8,
    ...     extra_node_props={"label": custom_labels, "score": custom_scores,
    ...         "confidence": "float64"},
    ...     extra_edge_props={"weight": custom_edge_weights, "type": "str"},
    ...     include_t=True,
    ...     include_z=False,
    ...     include_y=True,
    ...     include_x=True,
    ... )

    # To construct an graph with a backend graph lib
    >>> # Import construct function of your choice
    >>> store, memory_geff = create_simple_2d_geff()
    >>> graph = construct_nx(**memory_geff)
    >>> # graph is a networkx Graph ready for analysis
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, TypedDict, get_args

import numpy as np
import zarr
import zarr.storage

from geff.core_io import write_arrays
from geff_spec import Axis, PropMetadata
from geff_spec.utils import (
    add_or_update_props_metadata,
    create_or_update_metadata,
    create_props_metadata,
)

if TYPE_CHECKING:
    from collections.abc import Mapping

    from numpy.typing import NDArray

    from geff._typing import InMemoryGeff, PropDictNpArray
    from geff_spec._valid_values import AxisType, SpaceUnits, TimeUnits


DTypeStr = Literal["double", "int", "int8", "uint8", "int16", "uint16", "float32", "float64", "str"]
NodeIdDTypeStr = Literal["uint", "uint8", "uint16", "uint32", "uint64"]
AxesOptions = Literal["t", "z", "y", "x"]


class ExampleNodeAxisPropsDtypes(TypedDict):
    position: DTypeStr
    time: DTypeStr


def create_dummy_in_mem_geff(
    node_id_dtype: NodeIdDTypeStr,
    node_axis_dtypes: ExampleNodeAxisPropsDtypes,
    directed: bool,
    num_nodes: int = 5,
    num_edges: int = 4,
    extra_node_props: Mapping[str, DTypeStr | NDArray[Any]] | None = None,
    extra_edge_props: Mapping[str, DTypeStr | NDArray[Any]] | None = None,
    include_t: bool = True,
    include_z: bool = True,
    include_y: bool = True,
    include_x: bool = True,
    include_varlength: bool = False,
    include_missing: bool = False,
) -> InMemoryGeff:
    """Create dummy graph properties for testing.

    Args:
        node_id_dtype: Data type for node IDs
        node_axis_dtypes: Dictionary specifying dtypes for node axis properties (space and time)
        directed: Whether the graph is directed
        num_nodes: Number of nodes to generate
        num_edges: Number of edges to generate
        extra_node_props: Dict mapping property names to dtypes for extra node properties
        extra_edge_props: Dict mapping property names to dtypes for extra edge properties
        include_t: Whether to include time dimension
        include_z: Whether to include z dimension
        include_y: Whether to include y dimension
        include_x: Whether to include x dimension
        include_varlength: Whether to include a variable length property. If true, will
            make a property on nodes called "var_length" that has 2d np arrays of various
            shapes
        include_missing: If true, creades a node and edge prop called "sparse_prop" where every
            other node/edge has a missing value

    Returns:
        InMemoryGeff containing all graph properties
    """
    # Generate nodes with flexible count
    nodes = np.arange(num_nodes, dtype=node_id_dtype)
    node_props: dict[str, PropDictNpArray] = {}
    node_prop_meta: list[PropMetadata] = []

    # Generate spatiotemporal coordinates with flexible dimensions
    def _add_axis(
        name: str,
        ax_type: Literal[AxisType],
        unit: str | Literal[SpaceUnits] | Literal[TimeUnits],
        values: np.ndarray,
    ) -> PropMetadata:
        node_props[name] = {"values": values, "missing": None}
        if num_nodes > 0:
            roimin, roimax = values.min(), values.max()
        else:
            roimin, roimax = None, None
        axes.append(Axis(name=name, type=ax_type, unit=unit, min=roimin, max=roimax))
        return create_props_metadata(identifier=name, prop_data=node_props[name], unit=unit)

    axes: list[Axis] = []
    if include_t:
        meta = _add_axis(
            name="t",
            ax_type="time",
            unit="second",
            values=np.array(
                [(i * 5 // num_nodes) + 1 for i in range(num_nodes)],
                dtype=node_axis_dtypes["time"],
            ),
        )
        node_prop_meta.append(meta)
    if include_z:
        meta = _add_axis(
            name="z",
            ax_type="space",
            unit="nanometer",
            values=np.linspace(0.5, 0.1, num_nodes, dtype=node_axis_dtypes["position"]),
        )
        node_prop_meta.append(meta)
    if include_y:
        meta = _add_axis(
            name="y",
            ax_type="space",
            unit="nanometer",
            values=np.linspace(100.0, 500.0, num_nodes, dtype=node_axis_dtypes["position"]),
        )
        node_prop_meta.append(meta)
    if include_x:
        meta = _add_axis(
            name="x",
            ax_type="space",
            unit="nanometer",
            values=np.linspace(1.0, 0.1, num_nodes, dtype=node_axis_dtypes["position"]),
        )
        node_prop_meta.append(meta)

    # Generate edges with flexible count (ensure we don't exceed possible edges)
    max_possible_edges = (
        num_nodes * (num_nodes - 1) // 2 if not directed else num_nodes * (num_nodes - 1)
    )
    actual_num_edges = min(num_edges, max_possible_edges)

    # Create edges ensuring we don't create duplicates
    edges_: list[list[Any]] = []
    edge_count = 0

    # For undirected graphs, we need to be more careful about duplicates
    if not directed:
        # Create a simple chain first, then add cross edges
        for i in range(min(actual_num_edges, num_nodes - 1)):
            source_idx = i
            target_idx = i + 1
            edges_.append([int(source_idx), int(target_idx)])
            edge_count += 1

        # Add remaining edges as cross connections
        remaining_edges = actual_num_edges - edge_count
        for i in range(remaining_edges):
            source_idx = i % (num_nodes - 2)
            target_idx = (i + 2) % (num_nodes - 1) + 1
            if source_idx != target_idx:
                edges_.append([int(source_idx), int(target_idx)])
                edge_count += 1
    else:
        # For directed graphs, we can create more edges efficiently
        edges_ = []
        edge_count = 0
        created_edges = set()  # Track created edges to avoid duplicates

        # First create a chain of edges
        for i in range(min(actual_num_edges, num_nodes - 1)):
            source_idx = i
            target_idx = i + 1
            edge_tuple = (int(source_idx), int(target_idx))
            if edge_tuple not in created_edges:
                edges_.append([int(source_idx), int(target_idx)])
                created_edges.add(edge_tuple)
                edge_count += 1

        # Add remaining edges using different patterns
        remaining_edges = actual_num_edges - edge_count
        if remaining_edges > 0:
            # Create edges with different offsets
            for i in range(remaining_edges * 2):  # Try more iterations to find unique edges
                source_idx = i % num_nodes
                target_idx = (i + 2) % num_nodes  # Skip one node
                if source_idx != target_idx:
                    edge_tuple = (int(source_idx), int(target_idx))
                    if edge_tuple not in created_edges:
                        edges_.append([int(source_idx), int(target_idx)])
                        created_edges.add(edge_tuple)
                        edge_count += 1

                        # Stop if we've reached the target
                        if edge_count >= actual_num_edges:
                            break

            # If we still need more edges, use another pattern
            if edge_count < actual_num_edges:
                for i in range(actual_num_edges * 2):  # Try more iterations to find unique edges
                    source_idx = i % num_nodes
                    target_idx = (i + 3) % num_nodes  # Skip two nodes
                    if source_idx != target_idx:
                        edge_tuple = (int(source_idx), int(target_idx))
                        if edge_tuple not in created_edges:
                            edges_.append([int(source_idx), int(target_idx)])
                            created_edges.add(edge_tuple)
                            edge_count += 1

                            # Stop if we've reached the target
                            if edge_count >= actual_num_edges:
                                break

    edges = np.array(edges_, dtype=node_id_dtype)
    if edges.shape[0] == 0:
        edges = edges.reshape((0, 2))

    # Generate extra node properties
    if extra_node_props is not None:
        # Validate input is a dict
        if not isinstance(extra_node_props, dict):
            raise ValueError(f"extra_node_props must be a dict, got {type(extra_node_props)}")

        # Validate dict contains only string keys and valid dtype values or numpy arrays
        for prop_name, prop_value in extra_node_props.items():
            if not isinstance(prop_name, str):
                raise ValueError(f"extra_node_props keys must be strings, got {type(prop_name)}")

            # Check if value is a string (dtype) or numpy array
            if isinstance(prop_value, str):
                # Auto-generate array with specified dtype
                prop_dtype = prop_value

                # Validate dtype is supported using DTypeStr
                valid_dtypes = get_args(DTypeStr)
                if prop_dtype not in valid_dtypes:
                    raise ValueError(
                        f"extra_node_props[{prop_name}] dtype '{prop_dtype}' not supported. "
                        f"Valid dtypes: {valid_dtypes}"
                    )

                # Generate different patterns for different property types
                if prop_dtype == "str":
                    values = np.array(
                        [f"{prop_name}_{i}" for i in range(num_nodes)], dtype=prop_dtype
                    )
                elif prop_dtype in ["int", "int8", "uint8", "int16", "uint16"]:
                    values = np.arange(num_nodes, dtype=prop_dtype)
                else:  # float types
                    values = np.linspace(0.1, 1.0, num_nodes, dtype=prop_dtype)
                prop_dict: PropDictNpArray = {"values": values, "missing": None}
                node_props[prop_name] = prop_dict
                node_prop_meta.append(create_props_metadata(prop_name, prop_dict))

            elif isinstance(prop_value, np.ndarray):
                # Use provided array directly
                # Validate array length matches num_nodes
                if len(prop_value) != num_nodes:
                    raise ValueError(
                        f"extra_node_props[{prop_name}] array length {len(prop_value)} "
                        f"does not match num_nodes {num_nodes}"
                    )

                node_props[prop_name] = {"values": prop_value, "missing": None}

            else:
                raise ValueError(
                    f"extra_node_props[{prop_name}] must be a string dtype or numpy array, "
                    f"got {type(prop_value)}"
                )

    # Generate edge properties
    edge_props_dict: dict[str, PropDictNpArray] = {}
    edge_prop_meta: list[PropMetadata] = []
    # Generate edge properties from extra_edge_props
    if extra_edge_props is not None:
        # Validate input is a dict
        if not isinstance(extra_edge_props, dict):
            raise ValueError(f"extra_edge_props must be a dict, got {type(extra_edge_props)}")

        # Validate dict contains only string keys and valid dtype values or numpy arrays
        for prop_name, prop_value in extra_edge_props.items():
            if not isinstance(prop_name, str):
                raise ValueError(f"extra_edge_props keys must be strings, got {type(prop_name)}")

            # Check if value is a string (dtype) or numpy array
            if isinstance(prop_value, str):
                # Auto-generate array with specified dtype
                prop_dtype = prop_value

                # Validate dtype is supported using DTypeStr
                valid_dtypes = get_args(DTypeStr)
                if prop_dtype not in valid_dtypes:
                    raise ValueError(
                        f"extra_edge_props[{prop_name}] dtype '{prop_dtype}' not supported. "
                        f"Valid dtypes: {valid_dtypes}"
                    )

                # Generate different patterns for different property types
                if prop_dtype == "str":
                    values = np.array(
                        [f"{prop_name}_{i}" for i in range(len(edges))], dtype=prop_dtype
                    )
                elif prop_dtype in ["int", "int8", "uint8", "int16", "uint16"]:
                    values = np.arange(len(edges), dtype=prop_dtype)
                else:  # float types
                    values = np.linspace(0.1, 1.0, len(edges), dtype=prop_dtype)

                edge_props_dict[prop_name] = {"values": values, "missing": None}
            elif isinstance(prop_value, np.ndarray):
                # Use provided array directly
                # Validate array length matches num_edges
                if len(prop_value) != len(edges):
                    raise ValueError(
                        f"extra_edge_props[{prop_name}] array length {len(prop_value)} "
                        f"does not match number of edges {len(edges)}"
                    )

                edge_props_dict[prop_name] = {"values": prop_value, "missing": None}

            else:
                raise ValueError(
                    f"extra_edge_props[{prop_name}] must be a string dtype or numpy array, "
                    f"got {type(prop_value)}"
                )

            edge_prop_meta.append(
                create_props_metadata(identifier=prop_name, prop_data=edge_props_dict[prop_name])
            )

    if include_varlength:
        prop_name = "var_length"
        ndim = 3
        _dtype = np.uint64
        values_list = []
        for node in range(num_nodes):
            shape = [
                node,
            ] * ndim
            arr = np.ones(shape=shape, dtype=_dtype) * node
            values_list.append(arr)

        values = np.array(values_list, dtype=np.object_)
        missing = np.zeros(shape=(num_nodes,), dtype=np.bool_)
        if num_nodes > 0:
            missing[0] = 1
        prop_dict = {"values": values, "missing": missing}
        node_props[prop_name] = prop_dict
        node_prop_meta.append(create_props_metadata(prop_name, prop_dict))

    if include_missing:
        prop_name = "sparse_prop"
        values = np.arange(num_nodes, dtype="float64")
        missing = np.zeros(num_nodes, dtype=np.bool_)
        if num_nodes > 0:
            missing[::2] = 1
        prop_dict = {"values": values, "missing": missing}
        prop_meta = create_props_metadata(prop_name, prop_dict)
        node_props[prop_name] = prop_dict
        node_prop_meta.append(prop_meta)
        edge_props_dict[prop_name] = prop_dict
        edge_prop_meta.append(prop_meta)

    metadata = create_or_update_metadata(metadata=None, is_directed=directed, axes=axes)
    metadata = add_or_update_props_metadata(metadata, node_prop_meta, "node")
    metadata = add_or_update_props_metadata(metadata, edge_prop_meta, "edge")

    return {
        "metadata": metadata,
        "node_ids": nodes,
        "edge_ids": edges,
        "node_props": node_props,
        "edge_props": edge_props_dict,
    }


def create_mock_geff(
    node_id_dtype: NodeIdDTypeStr,
    node_axis_dtypes: ExampleNodeAxisPropsDtypes,
    directed: bool,
    num_nodes: int = 5,
    num_edges: int = 4,
    extra_node_props: Mapping[str, DTypeStr | NDArray[Any]] | None = None,
    extra_edge_props: Mapping[str, DTypeStr | NDArray[Any]] | None = None,
    include_t: bool = True,
    include_z: bool = True,
    include_y: bool = True,
    include_x: bool = True,
    include_varlength: bool = False,
    include_missing: bool = False,
) -> tuple[zarr.storage.MemoryStore, InMemoryGeff]:
    """Create a mock geff in memory and return the zarr store and the in memory geff.

    Args:
        node_id_dtype: Data type for node IDs
        node_axis_dtypes: Dictionary specifying dtypes for node axis properties (space and time)
        directed: Whether the graph is directed
        num_nodes: Number of nodes to generate
        num_edges: Number of edges to generate
        extra_node_props: Dict mapping property names to dtypes for extra node properties
        extra_edge_props: Dict mapping property names to dtypes for extra edge properties
        include_t: Whether to include time dimension
        include_z: Whether to include z dimension
        include_y: Whether to include y dimension
        include_x: Whether to include x dimension
        include_varlength: Whether to include a variable length property. If true, will
            make a property on nodes called "var_length" that has 2d np arrays of various
            shapes
        include_missing: If true, creades a node prop called "sparse_prop" where every other
            node has a missing value

    Returns:
        Tuple of (zarr store in memory, InMemoryGeff)
    """
    memory_geff = create_dummy_in_mem_geff(
        node_id_dtype=node_id_dtype,
        node_axis_dtypes=node_axis_dtypes,
        directed=directed,
        num_nodes=num_nodes,
        num_edges=num_edges,
        extra_node_props=extra_node_props,
        extra_edge_props=extra_edge_props,
        include_t=include_t,
        include_z=include_z,
        include_y=include_y,
        include_x=include_x,
        include_varlength=include_varlength,
    )

    # Create memory store and write graph to it
    store = zarr.storage.MemoryStore()
    write_arrays(store, **memory_geff)

    return store, memory_geff


def create_simple_2d_geff(
    num_nodes: int = 10,
    num_edges: int = 15,
    directed: bool = False,
) -> tuple[zarr.storage.MemoryStore, InMemoryGeff]:
    """Create a simple 2D geff graph with default settings.

    This is a convenience function for creating basic 2D graphs without having to
    specify all the detailed parameters. Uses sensible defaults for common use cases.

    Args:
        num_nodes: Number of nodes to generate (default: 10)
        num_edges: Number of edges to generate (default: 15)
        directed: Whether the graph is directed (default: False)

    Returns:
        Tuple of (zarr store in memory, InMemoryGeff)
    """
    return create_mock_geff(
        node_id_dtype="uint",
        node_axis_dtypes={"position": "float64", "time": "float64"},
        directed=directed,
        num_nodes=num_nodes,
        num_edges=num_edges,
        extra_edge_props={"score": "float64", "color": "int"},
        include_t=True,
        include_z=False,  # 2D only
        include_y=True,
        include_x=True,
    )


def create_simple_3d_geff(
    num_nodes: int = 10,
    num_edges: int = 15,
    directed: bool = False,
) -> tuple[zarr.storage.MemoryStore, InMemoryGeff]:
    """Create a simple 3D geff graph with default settings.

    This is a convenience function for creating basic 3D graphs without having to
    specify all the detailed parameters. Uses sensible defaults for common use cases.

    Args:
        num_nodes: Number of nodes to generate (default: 10)
        num_edges: Number of edges to generate (default: 15)
        directed: Whether the graph is directed (default: False)

    Returns:
        Tuple of (zarr store in memory, InMemoryGeff)
    """
    return create_mock_geff(
        node_id_dtype="uint",
        node_axis_dtypes={"position": "float64", "time": "float64"},
        directed=directed,
        num_nodes=num_nodes,
        num_edges=num_edges,
        extra_edge_props={"score": "float64", "color": "int"},
        include_t=True,
        include_z=True,  # 3D includes z
        include_y=True,
        include_x=True,
    )


def create_simple_temporal_geff(
    num_nodes: int = 10,
    num_edges: int = 15,
    directed: bool = False,
) -> tuple[zarr.storage.MemoryStore, InMemoryGeff]:
    """Create a simple geff graph with only time dimension (no spatial dimensions).

    This function creates a graph with nodes, edges, and time coordinates,
    but no spatial dimensions (x, y, z). Useful for temporal-only analysis.

    Args:
        num_nodes: Number of nodes to generate (default: 10)
        num_edges: Number of edges to generate (default: 15)
        directed: Whether the graph is directed (default: False)

    Returns:
        Tuple of (zarr store in memory, InMemoryGeff)
    """
    return create_mock_geff(
        node_id_dtype="uint",
        node_axis_dtypes={"position": "float64", "time": "float64"},
        directed=directed,
        num_nodes=num_nodes,
        num_edges=num_edges,
        extra_edge_props={"score": "float64", "color": "int"},
        include_t=True,
        include_z=False,  # No spatial dimensions
        include_y=False,  # No spatial dimensions
        include_x=False,  # No spatial dimensions
    )


def create_empty_geff(directed: bool = False) -> tuple[zarr.storage.MemoryStore, InMemoryGeff]:
    """Creates a geff without any nodes or edges

    Args:
        directed (bool, optional): Whether to create a directed graph. Defaults to False.

    Returns:
        tuple[zarr.storage.MemoryStore, InMemoryGeff]
    """
    return create_mock_geff(
        node_id_dtype="uint",
        node_axis_dtypes={"position": "float64", "time": "float64"},
        directed=directed,
        num_nodes=0,
        num_edges=0,
        include_t=False,  # No time
        include_z=False,  # No spatial dimensions
        include_y=False,  # No spatial dimensions
        include_x=False,  # No spatial dimensions
    )



import numpy as np
import pytest

from geff.core_io import construct_var_len_props
from geff.core_io._serialization import (
    _deserialize_vlen_value,
    deserialize_vlen_property_data,
    serialize_vlen_property_data,
)


def basic_1d_data():
    """Basic 1D test data."""
    prop_dict = construct_var_len_props(
        [
            np.array([1, 2, 3]),
            np.array([4, 5]),
            np.array([6]),
        ]
    )
    offset_shapes = [
        [0, 3],
        [3, 2],
        [5, 1],
    ]
    serialized = np.array([1, 2, 3, 4, 5, 6])
    serialized_dict = {
        "values": np.array(offset_shapes, dtype=np.uint64),
        "missing": None,
        "data": serialized,
    }
    return prop_dict, serialized_dict


def data_with_missing():
    prop_dict = {
        "values": np.array(
            [
                np.array([1, 2, 3]),
                np.array([4, 5]),
                np.array([6]),
            ],
            dtype=np.object_,
        ),
        "missing": np.array([0, 1, 0], dtype=np.bool_),
    }
    offset_shapes = [
        [0, 3],
        [3, 2],
        [5, 1],
    ]
    serialized = np.array([1, 2, 3, 4, 5, 6])
    serialized_dict = {
        "values": np.array(offset_shapes, dtype=np.uint64),
        "missing": np.array([0, 1, 0], dtype=np.bool_),
        "data": serialized,
    }
    return prop_dict, serialized_dict


def basic_2d_data():
    """Basic 2D test data."""
    prop_dict = construct_var_len_props([[[1, 2], [3, 4]], [[5, 6, 7], [8, 9, 10]]])
    offset_shapes = [
        [0, 2, 2],
        [4, 2, 3],
    ]
    serialized = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    serialized_dict = {
        "values": np.array(offset_shapes, dtype=np.uint64),
        "missing": None,
        "data": serialized,
    }
    return prop_dict, serialized_dict


def complex_2d_data():
    """Complex 2D test data with different shapes."""
    prop_dict = construct_var_len_props(
        [
            [[1, 2], [3, 4]],  # 2x2
            None,
            np.array([[5, 6, 7]]),  # 1x3
            np.array([[8], [9], [10]]),  # 3x1
        ]
    )
    offset_shapes = [
        [0, 2, 2],
        [4, 0, 0],
        [4, 1, 3],
        [7, 3, 1],
    ]
    serialized = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
    serialized_dict = {
        "values": np.array(offset_shapes, dtype=np.uint64),
        "missing": np.array([0, 1, 0, 0], dtype=np.bool_),
        "data": serialized,
    }
    return prop_dict, serialized_dict


def empty_and_data():
    """Test data with empty arrays."""
    prop_dict = {
        "values": np.array(
            [
                np.array([], dtype="int"),
                np.array([1, 2], dtype="int"),
                np.array([], dtype="int"),
            ],
            dtype=np.object_,
        ),
        "missing": None,
    }
    offset_shapes = [
        [
            0,
            0,
        ],
        [0, 2],
        [2, 0],
    ]
    serialized = np.array([1, 2])
    serialized_dict = {
        "values": np.array(offset_shapes, dtype=np.uint64),
        "missing": None,
        "data": serialized,
    }
    return prop_dict, serialized_dict


def all_missing_data():
    prop_dict = {
        "values": np.array(
            [
                np.array([], dtype="int"),
                np.array([1, 2], dtype="int"),
                np.array([], dtype="int"),
            ],
            dtype=np.object_,
        ),
        "missing": np.array(
            [
                1,
                1,
                1,
            ],
            dtype=np.bool_,
        ),
    }
    offset_shapes = [
        [
            0,
            0,
        ],
        [0, 2],
        [2, 0],
    ]
    serialized = np.array([1, 2])
    serialized_dict = {
        "values": np.array(offset_shapes, dtype=np.uint64),
        "missing": np.array(
            [
                1,
                1,
                1,
            ],
            dtype=np.bool_,
        ),
        "data": serialized,
    }
    return prop_dict, serialized_dict


def mixed_dimension_data():
    """Test data with mixed dimensions (should expand dims of the first array)."""
    prop_dict = construct_var_len_props([np.array([1, 2, 3]), np.array([[4, 5], [6, 7]])])

    offset_shapes = [
        [0, 1, 3],
        [3, 2, 2],
    ]
    serialized = np.array([1, 2, 3, 4, 5, 6, 7])
    serialized_dict = {
        "values": np.array(offset_shapes, dtype=np.uint64),
        "missing": None,
        "data": serialized,
    }
    return prop_dict, serialized_dict


def float_data():
    """Floating point test data."""
    prop_dict = construct_var_len_props(
        [
            np.array([1.1, 2.2, 3.3]),
            [],
            np.array([4.4]),
        ]
    )
    offset_shapes = [
        [
            0,
            3,
        ],
        [3, 0],
        [3, 1],
    ]
    serialized = np.array([1.1, 2.2, 3.3, 4.4])
    serialized_dict = {
        "values": np.array(offset_shapes, dtype=np.uint64),
        "missing": None,
        "data": serialized,
    }
    return prop_dict, serialized_dict


class TestSerializeVlenPropertyData:
    """Test suite for serialize_vlen_property_data function."""

    @pytest.mark.parametrize(
        "case",
        [
            basic_1d_data,
            data_with_missing,
            basic_2d_data,
            complex_2d_data,
            empty_and_data,
            all_missing_data,
            mixed_dimension_data,
            float_data,
        ],
    )
    def test_serialize_basic_data(self, case):
        """Test serialization of basic variable-length data."""
        prop_dict, ser_dict = case()
        values, missing, data = serialize_vlen_property_data(prop_dict)
        expected_values = ser_dict["values"]
        expected_missing = ser_dict["missing"]
        expected_data = ser_dict["data"]
        np.testing.assert_array_equal(values, expected_values)
        np.testing.assert_array_equal(missing, expected_missing)
        np.testing.assert_array_equal(data, expected_data)

    def test_serialize_nonarrays(self):
        with pytest.raises(
            ValueError,
            match="For variable length properties, each node/edge property must be a numpy array",
        ):
            serialize_vlen_property_data(
                {
                    "values": [[1, 2]],
                    "missing": None,
                }
            )

    def test_serialize_mixed_ndim(self):
        mixed_dimension_data = [np.array([1, 2, 3]), np.array([[4, 5], [6, 7]])]
        with pytest.raises(
            ValueError,
            match="For variable length properties, each node/edge property must have the "
            "same number of dimensions",
        ):
            serialize_vlen_property_data({"values": mixed_dimension_data, "missing": None})

    def test_serialize_mixed_dtype(self):
        mixed_dtype_data = [
            np.array([1, 2, 3], dtype=np.int64),
            np.array([[4, 5], [6, 7]], np.float64),
        ]
        with pytest.raises(
            ValueError,
            match="For variable length properties, each node/edge property must have the "
            "same number of dimensions",
        ):
            serialize_vlen_property_data({"values": mixed_dtype_data, "missing": None})


class TestDeserializeVlenPropertyData:
    """Test suite for deserialize_vlen_property_data function."""

    @pytest.mark.parametrize(
        "case",
        [
            basic_1d_data,
            data_with_missing,
            basic_2d_data,
            complex_2d_data,
            empty_and_data,
            all_missing_data,
            mixed_dimension_data,
            float_data,
        ],
    )
    def test_deserialize(self, case):
        expected_prop_dict, ser_dict = case()
        values = ser_dict["values"]
        missing = ser_dict["missing"]
        data = ser_dict["data"]
        prop_dict = deserialize_vlen_property_data(values, missing, data)
        val = prop_dict["values"]
        expected_val = expected_prop_dict["values"]
        assert val.shape == expected_val.shape
        assert val.dtype == expected_val.dtype
        for item, expected_item in zip(val, expected_val, strict=True):
            np.testing.assert_array_equal(item, expected_item)

        miss = prop_dict["missing"]
        expected_miss = expected_prop_dict["missing"]
        if miss is None or expected_miss is None:
            assert miss is None and expected_miss is None
        else:
            np.testing.assert_array_equal(miss, expected_miss)

    def test_deserialize_index_out_of_bounds(self):
        """Test deserialization with out of bounds index."""
        _, ser_dict = basic_1d_data()
        with pytest.raises(IndexError, match="Index 5 out of bounds"):
            _deserialize_vlen_value(values=ser_dict["values"], data=ser_dict["data"], index=5)



from typing import ClassVar

import numpy as np
import pytest

from geff.validate.shapes import validate_ellipsoid, validate_sphere
from geff_spec import Axis


class Test_validate_ellipsoid:
    axes_2d: ClassVar[list[Axis]] = [Axis(name="x", type="space"), Axis(name="y", type="space")]
    axes_3d: ClassVar[list[Axis]] = [
        Axis(name="x", type="space"),
        Axis(name="y", type="space"),
        Axis(name="z", type="space"),
    ]

    def test_axes(self):
        arr = np.ones((10, 2, 2))
        # Must provided axes
        with pytest.raises(
            ValueError, match="Must define space axes in order to have ellipsoid data"
        ):
            validate_ellipsoid(arr, None)

        # Axes must be spatial
        axes = [Axis(name="t", type="time"), Axis(name="c", type="channel")]
        with pytest.raises(
            ValueError, match="Must define space axes in order to have ellipsoid data"
        ):
            validate_ellipsoid(arr, axes)

    def test_square_matrix(self):
        arr = np.ones((10, 2, 5))
        with pytest.raises(
            ValueError, match="Spatial dimensions of covariance matrix must be equal"
        ):
            validate_ellipsoid(arr, self.axes_2d)

    def test_ndim(self):
        arr = np.ones((10, 2, 2))
        with pytest.raises(
            ValueError, match=r"Ellipsoid covariance matrix must have .* dimensions"
        ):
            validate_ellipsoid(arr, self.axes_3d)

    def test_symmetric(self):
        arr = np.ones((10, 2, 2))
        arr[:, 0, 1] = 0
        with pytest.raises(ValueError, match="Ellipsoid covariance matrices must be symmetric"):
            validate_ellipsoid(arr, self.axes_2d)

    def test_pos_def(self):
        arr = np.ones((10, 2, 2))
        with pytest.raises(
            ValueError, match="Ellipsoid covariance matrices must be positive-definite"
        ):
            validate_ellipsoid(arr, self.axes_2d)


def test_validate_sphere():
    # Not 1d
    with pytest.raises(ValueError, match="Sphere radius values must be 1D"):
        validate_sphere(np.ones((2, 2, 2)))

    # Not positive
    with pytest.raises(ValueError, match=r"Sphere radius values must be non-negative."):
        validate_sphere(np.full((2), fill_value=-1))



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypedDict

from typing_extensions import NotRequired

if TYPE_CHECKING:
    import numpy as np
    import zarr
    from numpy.typing import NDArray

    from geff_spec import GeffMetadata

# From python 3.11 TypeDicts can also inherit from Generic
# While python 3.10 is support two PropDicts for NDArray and zarr.Array are defined
#
# # implementation with Generic for python 3.11
# from typing import TypeVar, Generic
#
# # the typevar T_Array means that the arrays can either be numpy or zarr arrays
# T_Array = TypeVar("T_Array", bound=zarr.Array | NDArray)
#
# class PropDict(TypedDict, Generic[T_Array]):
#     values: T_Array
#     missing: NotRequired[T_Array]


class PropDictNpArray(TypedDict):
    """
    A prop dictionary which has the keys "values" and optionally "missing", stored as numpy arrays.

    "values" is a numpy array of any type, "missing" is a numpy array of bools.
    Variable length properties should have object dtype for their values array
    """

    values: NDArray[Any]
    missing: NDArray[np.bool_] | None


class ZarrPropDict(TypedDict):
    """
    A prop dictionary which has the keys "values" and optionally "missing" and "data",
    stored as zarr arrays.
    """

    values: zarr.Array
    missing: NotRequired[zarr.Array]
    data: NotRequired[zarr.Array]


# Intermediate dict format that can be constructed to different backend types
class InMemoryGeff(TypedDict):
    """
    Geff data loaded into memory as numpy arrays.
    """

    metadata: GeffMetadata
    node_ids: NDArray[Any]
    edge_ids: NDArray[Any]
    node_props: dict[str, PropDictNpArray]
    edge_props: dict[str, PropDictNpArray]



from typing import Any, Literal, Protocol, TypeVar

from numpy.typing import NDArray
from zarr.storage import StoreLike

from geff._typing import PropDictNpArray
from geff.core_io._base_read import read_to_memory
from geff.validate.data import ValidationConfig
from geff_spec import AxisType, GeffMetadata

from ._graph_adapter import GraphAdapter

T = TypeVar("T")


# NOTE: see _api_wrapper.py read/write/construct functions for docstrings
class Backend(Protocol[T]):
    """
    A protocol that acts as a namespace for functions that allow for backend interoperability.
    """

    @property
    def GRAPH_TYPES(self) -> tuple[type[T], ...]: ...

    @staticmethod
    def construct(
        metadata: GeffMetadata,
        node_ids: NDArray[Any],
        edge_ids: NDArray[Any],
        node_props: dict[str, PropDictNpArray],
        edge_props: dict[str, PropDictNpArray],
        *args: Any,
        **kwargs: Any,
    ) -> T: ...

    @classmethod
    def read(
        cls,
        store: StoreLike,
        structure_validation: bool = True,
        node_props: list[str] | None = None,
        edge_props: list[str] | None = None,
        data_validation: ValidationConfig | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> tuple[T, GeffMetadata]:
        in_memory_geff = read_to_memory(
            store, structure_validation, node_props, edge_props, data_validation
        )
        return cls.construct(**in_memory_geff, **kwargs), in_memory_geff["metadata"]

    @staticmethod
    def write(
        graph: T,
        store: StoreLike,
        metadata: GeffMetadata | None = None,
        axis_names: list[str] | None = None,
        axis_units: list[str | None] | None = None,
        axis_types: list[AxisType | None] | None = None,
        axis_scales: list[float | None] | None = None,
        scaled_units: list[str | None] | None = None,
        axis_offset: list[float | None] | None = None,
        zarr_format: Literal[2, 3] = 2,
        structure_validation: bool = True,
        *args: Any,
        **kwargs: Any,
    ) -> None: ...

    @staticmethod
    def graph_adapter(graph: T) -> GraphAdapter[T]:
        """
        Wrap a graph in a GraphAdapter for a unified API for accessing data.

        Args:
            graph (T): An instance of a supported graph object.

        Returns:
            GraphAdapter[T]
        """
        ...



from __future__ import annotations

import copy

import numpy as np
import pytest
import zarr
import zarr.storage

from geff import _path, validate_structure
from geff.core_io._base_write import write_arrays
from geff.core_io._utils import expect_group, open_storelike
from geff.testing._utils import check_equiv_geff
from geff.testing.data import (
    create_mock_geff,
    create_simple_2d_geff,
)
from geff.validate.structure import (
    _validate_axes_structure,
    _validate_edges_group,
    _validate_nodes_group,
    _validate_props_group,
)
from geff_spec import GeffMetadata, PropMetadata


@pytest.fixture
def z() -> zarr.Group:
    store, _ = create_mock_geff(
        node_id_dtype="uint",
        node_axis_dtypes={"position": "float64", "time": "float64"},
        directed=False,
        num_nodes=10,
        num_edges=15,
        extra_node_props={"score": "float64"},
        extra_edge_props={"score": "float64", "color": "int"},
        include_t=True,
        include_z=True,  # 3D includes z
        include_y=True,
        include_x=True,
        include_varlength=True,
    )
    return zarr.open_group(store)


@pytest.fixture
def meta(z) -> GeffMetadata:
    return GeffMetadata.read(z.store)


@pytest.fixture
def node_group(z) -> zarr.Group:
    return expect_group(z, _path.NODES)


@pytest.fixture
def edge_group(z) -> zarr.Group:
    return expect_group(z, _path.EDGES)


class TestValidateStructure:
    def test_valid_geff(self, z):
        validate_structure(z.store)

    def test_missing_metadata(self, z):
        del z.attrs["geff"]

        # Missing metadata
        with pytest.raises(ValueError, match="No geff key found in"):
            validate_structure(z.store)

    def test_no_nodes_group(self, z):
        del z[_path.NODES]
        with pytest.raises(
            ValueError, match=f"'graph' group must contain a group named '{_path.NODES}'"
        ):
            validate_structure(z.store)

    def test_no_edges(self, z):
        del z["edges"]
        with pytest.raises(
            ValueError, match=f"'graph' group must contain a group named '{_path.EDGES}'"
        ):
            validate_structure(z.store)

    # other cases are tested in validate_nodes_group and validate_edges_group


class Test_validate_nodes_group:
    def test_no_node_ids(self, node_group, meta):
        del node_group[_path.IDS]
        with pytest.raises(
            ValueError, match=f"'{_path.NODES}' group must contain an '{_path.IDS}' array"
        ):
            _validate_nodes_group(node_group, meta)

    def test_no_node_props_group(self, node_group, meta):
        del node_group[_path.PROPS]
        # Nodes must have a props group
        with pytest.raises(
            ValueError, match=f"'{_path.NODES}' group must contain a group named '{_path.PROPS}'"
        ):
            _validate_nodes_group(node_group, meta)

    def test_ids_not_int(self, node_group, meta):
        node_group[_path.IDS] = node_group[_path.IDS][:].astype("float")
        with pytest.raises(ValueError, match="Node ids must have an integer dtype"):
            _validate_nodes_group(node_group, meta)

    # Other cases are caught in tests for _validate_props_group


class Test_validate_edges_group:
    def test_no_edge_ids(self, edge_group, meta):
        del edge_group[_path.IDS]
        with pytest.raises(
            ValueError, match=f"'{_path.EDGES}' group must contain an '{_path.IDS}' array"
        ):
            _validate_edges_group(edge_group, meta)

    def test_edge_ids_bad_second_dim(self, edge_group, meta):
        edge_group[_path.IDS] = np.zeros((3, 3))
        with pytest.raises(
            ValueError,
            match=r"edges ids must be 2d with last dimension of size 2, received shape .*",
        ):
            _validate_edges_group(edge_group, meta)

    def test_edge_ids_wrong_ndim(self, edge_group, meta):
        edge_group[_path.IDS] = np.zeros((3, 1))
        with pytest.raises(
            ValueError,
            match=r"edges ids must be 2d with last dimension of size 2, received shape .*",
        ):
            _validate_edges_group(edge_group, meta)

    def test_edge_ids_not_int(self, edge_group, meta):
        edge_group[_path.IDS] = edge_group[_path.IDS][:].astype("float")
        with pytest.raises(
            ValueError,
            match="Edge ids must have an integer dtype",
        ):
            _validate_edges_group(edge_group, meta)

        # int and uint are now both ok
        edge_group[_path.IDS] = edge_group[_path.IDS][:].astype("int")
        _validate_edges_group(edge_group, meta)

    # Other cases are caught in tests for _validate_props_group


class Test_validate_props_group:
    def test_node_prop_no_values(self, node_group, meta):
        # Subgroups in props must have values
        key = "t"
        del node_group[_path.PROPS][key][_path.VALUES]
        id_len = node_group[_path.IDS].shape[0]
        with pytest.raises(
            ValueError, match=f"Node property group '{key}' must have a '{_path.VALUES}' array"
        ):
            _validate_props_group(node_group[_path.PROPS], id_len, "Node", meta.node_props_metadata)

    def test_node_prop_shape_mismatch(self, node_group, meta):
        # Property shape mismatch
        key = "x"
        node_group[f"{_path.PROPS}/{key}/{_path.VALUES}"] = np.zeros(1)
        id_len = node_group[_path.IDS].shape[0]
        with pytest.raises(
            ValueError,
            match=(
                f"Node property '{key}' values has length {1}, which does not match id length .*"
            ),
        ):
            _validate_props_group(node_group[_path.PROPS], id_len, "Node", meta.node_props_metadata)

    def test_node_prop_missing_mismatch(self, node_group, meta):
        # Property missing shape mismatch
        key = "t"
        node_group[f"{_path.PROPS}/{key}/{_path.MISSING}"] = np.zeros(shape=(1))
        id_len = node_group[_path.IDS].shape[0]
        with pytest.raises(
            ValueError,
            match=(
                f"Node property '{key}' missing mask has length 1, "
                "which does not match id length .*"
            ),
        ):
            _validate_props_group(node_group[_path.PROPS], id_len, "Node", meta.node_props_metadata)

    def test_missing_dtype(self, node_group, meta):
        # missing arrays must be boolean
        key = "score"
        node_group[f"{_path.PROPS}/{key}/{_path.MISSING}"] = np.zeros(
            node_group[f"{_path.PROPS}/{key}/{_path.VALUES}"].shape, dtype="float"
        )
        id_len = node_group[_path.IDS].shape[0]

        with pytest.raises(ValueError, match=f"Node property '{key}' missing must be boolean"):
            _validate_props_group(node_group[_path.PROPS], id_len, "Node", meta.node_props_metadata)

    def test_missing_prop_metadata(self, node_group):
        id_len = node_group[_path.IDS].shape[0]
        with pytest.raises(ValueError, match=r"Property .* is missing from the property metadata"):
            _validate_props_group(node_group[_path.PROPS], id_len, "Node", {})

    def test_extra_prop_metadata(self, node_group, meta):
        meta.node_props_metadata["extra"] = PropMetadata(identifier="extra", dtype="int")
        id_len = node_group[_path.IDS].shape[0]
        with pytest.raises(
            ValueError, match=r"Property .* is in the metadata but missing from the property group"
        ):
            _validate_props_group(node_group[_path.PROPS], id_len, "Node", meta.node_props_metadata)

    def test_mismatched_dtype_prop_metadata(self, node_group, meta):
        id_len = node_group[_path.IDS].shape[0]
        x_meta = meta.node_props_metadata["x"]
        x_meta.dtype = "uint64"
        meta.node_props_metadata["x"] = x_meta

        with pytest.raises(
            ValueError, match=r"Property .* has stated dtype .* but actual dtype .*"
        ):
            _validate_props_group(node_group[_path.PROPS], id_len, "Node", meta.node_props_metadata)

    def test_varlen_value_dtype(self, node_group, meta):
        id_len = node_group[_path.IDS].shape[0]
        x_meta = meta.node_props_metadata["var_length"]
        x_meta.dtype = "float"
        meta.node_props_metadata["var_length"] = x_meta

        with pytest.raises(
            ValueError, match=r"Property .* has stated dtype .* but actual dtype .*"
        ):
            _validate_props_group(node_group[_path.PROPS], id_len, "Node", meta.node_props_metadata)

    def test_varlen_data_dtype(self, node_group, meta):
        id_len = node_group[_path.IDS].shape[0]
        node_group[_path.PROPS]["var_length"][_path.VALUES] = np.zeros(
            shape=(id_len, 2), dtype="float"
        )

        with pytest.raises(
            ValueError, match=r"Varlength property .* values array does not have type uint64"
        ):
            _validate_props_group(node_group[_path.PROPS], id_len, "Node", meta.node_props_metadata)

    def test_data_wo_varlen(self, node_group, meta):
        id_len = node_group[_path.IDS].shape[0]
        node_group[_path.PROPS]["x"][_path.DATA] = np.zeros(10)

        with pytest.raises(
            ValueError, match=r"Found data array for property .* which is not a varlength property"
        ):
            _validate_props_group(node_group[_path.PROPS], id_len, "Node", meta.node_props_metadata)


def test_open_storelike(tmp_path):
    # Open from a path
    valid_zarr = f"{tmp_path}/test.zarr"
    _ = zarr.open(valid_zarr)
    group = open_storelike(valid_zarr)
    assert isinstance(group, zarr.Group)

    # Open from a store
    store = zarr.storage.MemoryStore()
    zarr.open_group(store, path="group")
    group = open_storelike(store)
    assert isinstance(group, zarr.Group)

    # Bad path
    with pytest.raises(FileNotFoundError, match="Path does not exist"):
        open_storelike(f"{tmp_path}/bad.zarr")

    # Not a store
    with pytest.raises(ValueError, match="store must be a zarr StoreLike"):
        open_storelike(group)


class Test_check_equiv_geff:
    store, attrs = create_simple_2d_geff(num_nodes=10, num_edges=15)
    in_mem = attrs

    def _write_new_store(self, in_mem):
        store = zarr.storage.MemoryStore()
        write_arrays(store, **in_mem)
        return store

    def test_same_geff(self):
        # Check that two exactly same geffs pass
        check_equiv_geff(self.store, self.store)

    def test_id_shape_mismatch(self):
        # Id shape mismatch
        bad_store, _attrs = create_simple_2d_geff(num_nodes=5)
        with pytest.raises(ValueError, match=r".* ids shape: .* does not match .*"):
            check_equiv_geff(self.store, bad_store)

    def test_props_mismatch(self):
        bad_mem = copy.deepcopy(self.in_mem)
        bad_mem["node_props"]["new prop"] = bad_mem["node_props"]["t"]
        bad_mem["metadata"].node_props_metadata["new prop"] = PropMetadata(
            identifier="new prop", dtype="uint64"
        )
        bad_store = self._write_new_store(bad_mem)
        with pytest.raises(ValueError, match=r".* properties: a .* does not match b .*"):
            check_equiv_geff(self.store, bad_store)

    def test_only_one_with_missing(self):
        bad_mem = copy.deepcopy(self.in_mem)
        bad_mem["edge_props"]["score"]["missing"] = np.zeros(
            bad_mem["edge_props"]["score"]["values"].shape, dtype=np.bool_
        )
        bad_store = self._write_new_store(bad_mem)
        with pytest.raises(UserWarning, match=r".* contains missing but the other does not"):
            check_equiv_geff(bad_store, self.store)

    def test_value_shape_mismatch(self):
        bad_mem = copy.deepcopy(self.in_mem)
        # Add extra dimension to an edge prop
        bad_mem["edge_props"]["score"]["values"] = bad_mem["edge_props"]["score"]["values"][
            ..., np.newaxis
        ]
        bad_store = self._write_new_store(bad_mem)
        with pytest.raises(ValueError, match=r".* shape: .* does not match b .*"):
            check_equiv_geff(self.store, bad_store)

    def test_value_dtype_mismatch(self):
        # Values dtype mismatch
        bad_mem = copy.deepcopy(self.in_mem)
        # Change dtype
        bad_mem["edge_props"]["score"]["values"] = (
            bad_mem["edge_props"]["score"]["values"].astype("int").squeeze()
        )
        bad_store = self._write_new_store(bad_mem)
        with pytest.raises(ValueError, match=r".* dtype: .* does not match b .*"):
            check_equiv_geff(self.store, bad_store)


class Test_validate_axes_structure:
    def test_missing_axes_prop(self, z, meta):
        key = "x"
        del z[_path.NODE_PROPS][key]
        del meta.node_props_metadata["x"]
        with pytest.raises(ValueError, match=f"Axis {key} data is missing"):
            _validate_axes_structure(z, meta)

    def test_must_be_1d(self, z, meta):
        z[f"{_path.NODE_PROPS}/x/{_path.VALUES}"] = np.zeros((10, 2))
        with pytest.raises(ValueError, match="Axis property x has 2 dimensions, must be 1D"):
            _validate_axes_structure(z, meta)

    def test_no_missing_values(self, z, meta):
        z[f"{_path.NODE_PROPS}/x/{_path.VALUES}"] = np.zeros((10,))
        z[f"{_path.NODE_PROPS}/x/{_path.MISSING}"] = np.zeros((10,))
        with pytest.raises(ValueError, match="Axis x has missing values which are not allowed"):
            _validate_axes_structure(z, meta)



import os
from pathlib import Path

import numpy as np
import pytest
import zarr
import zarr.storage

from geff import _path
from geff.core_io._base_write import write_arrays
from geff.core_io._utils import (
    _detect_zarr_spec_version,
    check_for_geff,
    delete_geff,
    open_storelike,
    setup_zarr_group,
)
from geff.testing.data import create_simple_2d_geff


@pytest.mark.skipif(zarr.__version__.startswith("2"), reason="tests for zarr-python>=v3 only")
class TestZarrV3Warnings:
    """Test warnings for zarr version mismatches."""

    def test_detect_zarr_spec_version_v2(self, tmp_path: Path) -> None:
        """Test detection of zarr spec v2 files created explicitly."""
        zarr_path = tmp_path / "test_v2.zarr"

        # Create a zarr file explicitly using spec v2
        group = setup_zarr_group(str(zarr_path), zarr_format=2)
        group.create_array("test", shape=(10,), dtype=int)

        # Should detect as v2
        version = _detect_zarr_spec_version(str(zarr_path))
        assert version == 2

        # Should detect as v2
        version = _detect_zarr_spec_version(group.store)
        assert version == 2

    def test_detect_zarr_spec_version_v3(self, tmp_path: Path) -> None:
        """Test detection of zarr spec v3 files created explicitly."""
        zarr_path = tmp_path / "test_v3.zarr"

        # Create a zarr file explicitly using spec v3
        group = setup_zarr_group(str(zarr_path), zarr_format=3)
        group.create_array("test", shape=(10,), dtype=int)

        # Should detect as v3
        version = _detect_zarr_spec_version(str(zarr_path))
        assert version == 3

        # Should detect as v3
        version = _detect_zarr_spec_version(group.store)
        assert version == 3

    def test_open_storelike_warns_for_v3_with_zarr_v2(self, tmp_path: Path, monkeypatch) -> None:
        """Test that opening zarr spec v2 files with zarr python v2 produces a warning.

        Note: This test mocks that we're using zarr-python v2.
        """
        zarr_path = tmp_path / "test_v2.zarr"

        # Create a real zarr v2 file
        group = setup_zarr_group(str(zarr_path), zarr_format=3)
        group.create_array("test", shape=(10,), dtype=int)

        # Mock that we're using zarr-python v2
        monkeypatch.setattr(zarr, "__version__", "2.17.0")

        # Opening should produce a warning
        with pytest.warns(
            UserWarning, match="Attempting to open a zarr spec v3 file with zarr-python v2"
        ):
            open_storelike(str(zarr_path))


def test_detect_zarr_spec_version_returns_none_for_invalid_path():
    """Test that _detect_zarr_spec_version returns None for invalid paths."""
    version = _detect_zarr_spec_version("/nonexistent/path")
    assert version is None


@pytest.mark.skipif(not zarr.__version__.startswith("2"), reason="tests for zarr python v2 only")
class TestZarrV2Warnings:
    """Test warnings for zarr version mismatches."""

    def test_detect_zarr_spec_version_v2(self, tmp_path: Path) -> None:
        """Test detection of zarr spec v2 files created explicitly."""
        zarr_path = tmp_path / "test_v2.zarr"

        # Create a zarr file explicitly using spec v2
        group = setup_zarr_group(str(zarr_path), zarr_format=2)
        group.create_dataset("test", shape=(10,), dtype=int)

        # Should detect as v2
        version = _detect_zarr_spec_version(str(zarr_path))
        assert version == 2

    def test_warn_if_writing_zarr_format_3(self, tmp_path: Path):
        with pytest.warns(UserWarning, match="Requesting zarr spec v3 with zarr-python v2"):
            setup_zarr_group(tmp_path / "test.zarr", zarr_format=3)


class Test_delete_geff:
    def test_basic_path(self, tmp_path):
        _, mem_geff = create_simple_2d_geff()
        path = tmp_path / "test.geff"
        write_arrays(path, **mem_geff)
        assert os.path.exists(path)

        delete_geff(path)

        assert not os.path.exists(path)

    def test_extra_data(self, tmp_path):
        _, mem_geff = create_simple_2d_geff()
        path = tmp_path / "test.geff"
        write_arrays(path, **mem_geff, zarr_format=2)
        assert os.path.exists(path)
        # Test with extra data
        root = zarr.open(path, mode="a")
        root["other_data"] = np.zeros(shape=(10, 10))

        with pytest.raises(
            UserWarning,
            match=r"Found non-geff members in zarr. Exiting without deleting root zarr.",
        ):
            delete_geff(path)
            root = zarr.open(path)
            assert _path.NODES not in root
            assert _path.EDGES not in root
            assert "geff" not in root.attrs

    def test_mem_store(self):
        store, _ = create_simple_2d_geff()

        with pytest.raises(
            UserWarning,
            match="Cannot delete root zarr directory, but geff contents have been deleted",
        ):
            delete_geff(store)
            root = zarr.open(store)
            assert "geff" not in root.attrs


class Test_check_for_geff:
    def test_path(self, tmp_path):
        geff_path = tmp_path / "test.geff"
        # does not exist
        assert check_for_geff(geff_path) is False
        # exists
        os.mkdir(geff_path)
        assert check_for_geff(geff_path) is True

    def test_str(self, tmp_path):
        geff_path = str(tmp_path / "test.geff")
        # does not exist
        assert check_for_geff(geff_path) is False
        # exists
        os.mkdir(geff_path)
        assert check_for_geff(geff_path) is True

    def test_Store(self):
        store = zarr.storage.MemoryStore()
        # does not exist
        assert check_for_geff(store) is False
        # exists
        root = zarr.open(store)
        root.attrs["geff"] = "metadata"
        assert check_for_geff(store) is True



from __future__ import annotations

import sys

import pytest

# only run this file if benchmarks are requested, or running directly
if all(
    x not in {"--codspeed", "--benchmark", "packages/geff/tests/test_bench.py"} for x in sys.argv
):
    pytest.skip("use --benchmark to run benchmark", allow_module_level=True)

import atexit
import shutil
import tempfile
from functools import cache
from pathlib import Path
from types import MappingProxyType
from typing import TYPE_CHECKING, Any, get_args

import networkx as nx
import numpy as np

import geff
from geff._graph_libs._api_wrapper import SupportedBackend, get_backend

if TYPE_CHECKING:
    from collections.abc import Mapping

    from pytest_codspeed.plugin import BenchmarkFixture


np.random.seed(42)  # for reproducibility

# ###########################   Utils   ##################################

BACKEND_STRINGS: tuple[SupportedBackend] = get_args(SupportedBackend)


@cache
def node_data(n_nodes: int) -> Mapping[int, dict[str, float]]:
    """Returns a dict of {node_id -> tzyx_coord_dict}."""
    coords = np.random.uniform(size=(n_nodes, 4))
    nodes = {n: dict(zip("tzyx", c, strict=True)) for n, c in enumerate(coords)}
    return MappingProxyType(nodes)


@cache
def edge_data(n_nodes: int) -> Mapping[tuple[int, int], dict[str, Any]]:
    """Returns a dict of {(u, v) -> edge_data_dict}."""
    idx = np.arange(n_nodes)  # [0, 1, ..., n-1]
    u = np.repeat(idx, n_nodes)  # 0 0 ... 1 1 ...
    v = np.tile(idx, n_nodes)  # 0 1 ... 0 1 ...
    mask = u != v  # drop self-loops
    mask_sum = np.sum(mask)  # number of edges without self-loops
    edges = {
        (int(uu), int(vv)): {"float_prop": float(fp), "int_prop": int(ip)}
        for (uu, vv, fp, ip) in zip(
            u[mask],
            v[mask],
            np.random.uniform(size=mask_sum),
            np.arange(mask_sum, dtype=int),
            strict=True,
        )
    }
    return MappingProxyType(edges)


def create_nx_graph(num_nodes: int) -> nx.DiGraph:
    graph: nx.DiGraph[int] = nx.DiGraph()
    nodes, edges = node_data(num_nodes), edge_data(num_nodes)
    graph.add_nodes_from(nodes.items())
    graph.add_edges_from(((u, v, dd) for (u, v), dd in edges.items()))
    return graph


@cache
def graph_file_path(num_nodes: int) -> Path:
    tmp_dir = tempfile.mkdtemp(suffix=".zarr")
    atexit.register(shutil.rmtree, tmp_dir, ignore_errors=True)
    geff.write(
        graph=create_nx_graph(num_nodes),
        store=tmp_dir,
        axis_names=["t", "z", "y", "x"],
        overwrite=True,
    )
    return Path(tmp_dir)


# ###########################   TESTS   ##################################

# to keep consistency before api refac
WRITE_TEST_IDS: dict[SupportedBackend, str] = {
    "networkx": "write_nx",
    "rustworkx": "write_rx",
    "spatial-graph": "write_sg",
}


@pytest.mark.parametrize("nodes", [500])
@pytest.mark.parametrize(
    "backend",
    BACKEND_STRINGS,
    # to keep consistency before api refac
    ids=[WRITE_TEST_IDS[backend] for backend in BACKEND_STRINGS],
)
def test_bench_write(
    backend: SupportedBackend, benchmark: BenchmarkFixture, tmp_path: Path, nodes: int
) -> None:
    path = tmp_path / "test_write.zarr"
    backend_io = get_backend(backend)
    write_func = backend_io.write
    read_func = backend_io.read
    graph = read_func(graph_file_path(nodes))[0]
    benchmark.pedantic(
        write_func,
        kwargs={"graph": graph, "axis_names": ["t", "z", "y", "x"], "store": path},
        setup=lambda **__: shutil.rmtree(path, ignore_errors=True),  # delete previous zarr
    )


@pytest.mark.parametrize("nodes", [500])
def test_bench_validate(benchmark: BenchmarkFixture, nodes: int) -> None:
    graph_path = graph_file_path(nodes)
    benchmark(geff.validate_structure, store=graph_path)


# to keep consistency before api refac
READ_TEST_IDS: dict[SupportedBackend, str] = {
    "networkx": "read_nx",
    "rustworkx": "read_rx",
    "spatial-graph": "read_sg",
}


@pytest.mark.parametrize("nodes", [500])
@pytest.mark.parametrize(
    "backend",
    BACKEND_STRINGS,
    # to keep consistency before api refac
    ids=[READ_TEST_IDS[backend] for backend in BACKEND_STRINGS],
)
def test_bench_read(backend: SupportedBackend, benchmark: BenchmarkFixture, nodes: int) -> None:
    read_func = get_backend(backend).read
    graph_path = graph_file_path(nodes)
    benchmark(read_func, graph_path, structure_validation=False)



from __future__ import annotations

from pathlib import Path

import pytest
from typer.testing import CliRunner

from geff import GeffMetadata
from geff._cli import app
from geff.core_io._base_write import write_arrays
from geff.testing.data import create_simple_temporal_geff
from tests.test_convert.test_ctc import create_mock_data

runner = CliRunner()
TEST_DATA = Path(__file__).parent / "data"


@pytest.fixture
def example_geff_path(tmp_path: Path) -> str:
    file_path = str(tmp_path / "test.geff")
    _store, memory_geff = create_simple_temporal_geff()
    write_arrays(file_path, **memory_geff)
    return file_path


def test_validate_command_prints_valid(example_geff_path: str) -> None:
    """Test that the validate command prints the expected output."""
    result = runner.invoke(app, ["validate", example_geff_path])
    assert result.exit_code == 0


def test_info_command_prints_metadata(example_geff_path: str) -> None:
    result = runner.invoke(app, ["info", example_geff_path])
    metadata = GeffMetadata.read(example_geff_path)
    assert result.exit_code == 0
    assert result.output == metadata.model_dump_json(indent=2) + "\n"


def test_main_invalid_command() -> None:
    result = runner.invoke(app, ["invalid"])
    assert result.exit_code != 0


@pytest.mark.parametrize("is_gt", [True, False])
@pytest.mark.parametrize("tczyx", [True, False])
def test_convert_ctc(tmp_path: Path, is_gt: bool, tczyx: bool) -> None:
    ctc_path = create_mock_data(tmp_path, is_gt)
    geff_path = ctc_path / "little.geff"
    segm_path = ctc_path / "segm.zarr"

    cmd_args = [
        "convert-ctc",
        str(ctc_path),
        str(geff_path),
        "--segm-path",
        str(segm_path),
        "--input-image-dir",
        str(ctc_path),
        "--output-image-path",
        str(tmp_path / "output_image.zarr"),
    ]
    if tczyx:
        cmd_args.append("--tczyx")
    result = CliRunner().invoke(app, cmd_args)
    assert result.exit_code == 0, (
        f"{cmd_args} failed with exit code {result.exit_code} and message:\n{result.stdout}"
    )


@pytest.mark.parametrize(
    "other_arg", [None, "--discard-filtered-spots", "--discard-filtered-tracks", "--overwrite"]
)
def test_convert_trackmate_xml(tmp_path: Path, other_arg: str | None) -> None:
    geff_output = str(tmp_path / "test.geff")
    in_data = str(TEST_DATA / "FakeTracks.xml")

    cmd_args = ["convert-trackmate-xml", in_data, geff_output]
    if other_arg is not None:
        cmd_args.append(other_arg)

    with pytest.warns() as warning_list:
        result = CliRunner().invoke(app, cmd_args)
    warning_messages = [str(warning.message) for warning in warning_list]
    assert any("node properties were removed from the metadata" in msg for msg in warning_messages)
    assert any("edge property was removed from the metadata" in msg for msg in warning_messages)
    assert result.exit_code == 0, (
        f"{cmd_args} failed with exit code {result.exit_code} and message:\n{result.stdout}"
    )


def test_convert_to_csv(example_geff_path, tmp_path):
    out_path = tmp_path / "dataframe"
    result = CliRunner().invoke(app, ["convert-to-csv", str(example_geff_path), str(out_path)])
    assert result.exit_code == 0



from ._base_read import read_to_memory
from ._base_write import write_arrays, write_dicts
from ._utils import check_for_geff, construct_var_len_props, delete_geff

__all__ = [
    "check_for_geff",
    "construct_var_len_props",
    "delete_geff",
    "read_to_memory",
    "write_arrays",
    "write_dicts",
]



import numpy as np
import pytest

from geff.validate.tracks import validate_lineages, validate_tracklets


@pytest.mark.parametrize(
    "node_ids, edge_ids, tracklet_ids, expected_valid, description",
    [
        # Single, simple, valid tracklet (1→2→3)
        (
            np.array([1, 2, 3]),
            np.array([[1, 2], [2, 3]]),
            np.array([10, 10, 10]),
            True,
            "Valid simple path",
        ),
        # Tracklet with missing edge
        (
            np.array([1, 2, 3]),
            np.array([[1, 2]]),
            np.array([10, 10, 10]),
            False,
            "Missing edge in path",
        ),
        # Tracklet with a cycle
        (
            np.array([1, 2, 3]),
            np.array([[1, 2], [2, 3], [3, 1]]),
            np.array([10, 10, 10]),
            False,
            "Cycle in tracklet",
        ),
        # Multiple valid tracklets
        (
            np.array([1, 2, 3, 4, 5, 6]),
            np.array([[1, 2], [2, 3], [4, 5], [5, 6]]),
            np.array([10, 10, 10, 20, 20, 20]),
            True,
            "Two valid tracklets",
        ),
        # Branching in tracklet
        (
            np.array([1, 2, 3]),
            np.array([[1, 2], [1, 3]]),
            np.array([10, 10, 10]),
            False,
            "Branch in tracklet",
        ),
        # Not fully connected
        (
            np.array([1, 2, 3]),
            np.array([[1, 2]]),
            np.array([10, 10, 10]),
            False,
            "Not fully connected",
        ),
        # Two nodes, valid path
        (np.array([1, 2]), np.array([[1, 2]]), np.array([10, 10]), True, "Two nodes, valid path"),
        # Tracklet with all nodes, but disconnected
        (
            np.array([1, 2, 3, 4]),
            np.array([[1, 2], [3, 4]]),
            np.array([10, 10, 10, 10]),
            False,
            "Disconnected tracklet",
        ),
        # Multiple tracklets, one valid, one invalid
        (
            np.array([1, 2, 3, 4, 5, 6]),
            np.array([[1, 2], [2, 3], [4, 5]]),
            np.array([10, 10, 10, 20, 20, 20]),
            False,
            "One valid, one invalid",
        ),
        # Not maximal length tracklet
        (
            np.array([1, 2, 3, 4, 5]),
            np.array([[1, 2], [2, 3], [3, 4], [4, 5]]),
            np.array([10, 10, 10, 20, 20]),
            False,
            "Tracklet not maximal length",
        ),
    ],
)
def test_validate_tracklets(node_ids, edge_ids, tracklet_ids, expected_valid, description) -> None:
    is_valid, errors = validate_tracklets(node_ids, edge_ids, tracklet_ids)
    assert is_valid == expected_valid, f"{description} failed: {errors}"


@pytest.mark.parametrize(
    "node_ids, edge_ids, lineage_ids, expected_valid, description",
    [
        # --- Valid Cases (Lineage is a connected component) ---
        (
            np.array([1, 2, 3]),
            np.array([[1, 2], [2, 3]]),
            np.array([10, 10, 10]),
            True,
            "Valid: Simple connected lineage",
        ),
        (
            np.array([1, 2, 3]),
            np.array([[1, 2], [2, 3], [3, 1]]),
            np.array([10, 10, 10]),
            True,
            "Valid: A cycle connects all nodes",
        ),
        (
            np.array([1, 2, 3]),
            np.array([[1, 2], [1, 3]]),
            np.array([10, 10, 10]),
            True,
            "Valid: A branch connects all nodes",
        ),
        (
            np.array([1, 2, 3, 4, 5, 6]),
            np.array([[1, 2], [2, 3], [4, 5], [5, 6]]),
            np.array([10, 10, 10, 20, 20, 20]),
            True,
            "Valid: Two separate, fully-connected lineages",
        ),
        (
            np.array([1]),
            np.array([]),
            np.array([10]),
            True,
            "Valid: A single-node lineage is always a connected component",
        ),
        # --- Invalid Cases (Lineage is NOT a connected component) ---
        (
            np.array([1, 2, 3, 4]),
            np.array([[0, 1], [1, 2], [2, 3], [3, 4]]),
            np.array([10, 10, 10, 10]),
            False,
            "Invalid: Lineage is not isolated (part of a larger component)",
        ),
        (
            np.array([1, 2, 3]),
            np.array([[1, 2]]),
            np.array([10, 10, 10]),
            False,
            "Invalid: Lineage is not internally connected (node 3 is separate)",
        ),
        (
            np.array([1, 2, 3, 4]),
            np.array([[1, 2], [3, 4]]),
            np.array([10, 10, 10, 10]),
            False,
            "Invalid: Lineage contains two separate components",
        ),
        (
            np.array([1, 2, 3, 4, 5, 6]),
            np.array([[1, 2], [2, 3], [4, 5]]),
            np.array([10, 10, 10, 20, 20, 20]),
            False,
            "Invalid: One valid lineage, one invalid (disconnected) lineage",
        ),
        (
            np.array([1, 2, 3]),
            np.array([]),
            np.array([10, 10, 10]),
            False,
            "Invalid: Lineage contains three separate single-node components",
        ),
        (
            np.array([1, 2, 3]),
            np.array([[1, 2], [2, 3], [3, 4]]),
            np.array([10, 10, 10]),
            False,
            "Invalid: Lineage contains outside edges (node 4 not in lineage)",
        ),
    ],
)
def test_validate_lineages(node_ids, edge_ids, lineage_ids, expected_valid, description) -> None:
    """
    Tests the validate_lineages function for various connectivity scenarios.
    """
    is_valid, errors = validate_lineages(node_ids, edge_ids, lineage_ids)
    assert is_valid == expected_valid, f"Test '{description}' failed: {errors}"
